"""Implementations of MCMC Usher classes.

An usher is used to generate step proposals for MC Monte Carlo sampling.
For example a Flipper simply proposes a change of the identity of a species
at a site for use in a SemiGrand ensemble. A Swapper will propose a swap
between species at two sites for use in Canonical ensemble simulations.

More complex steps can be defined simply by deriving from the MCUsher.
"""

__author__ = "Luis Barroso-Luque, Fengyu Xie"

import warnings
from abc import ABC, abstractmethod

import numpy as np
from monty.json import jsanitize
from scipy.special import gammaln

from smol.moca.composition.space import CompositionSpace
from smol.moca.metadata import Metadata
from smol.moca.occu_utils import (
    delta_counts_from_step,
    get_dim_ids_by_sublattice,
    get_dim_ids_table,
    occu_to_counts,
    occu_to_species_list,
)
from smol.utils.class_utils import (
    class_name_from_str,
    derived_class_factory,
    get_subclasses_str,
)
from smol.utils.math import NUM_TOL, choose_section_from_partition, flip_weights_mask


class MCUsher(ABC):
    """Abstract base class for MC usher classes."""

    def __init__(self, sublattices, sublattice_probabilities=None, rng=None):
        """Initialize MCMCStep.

        Args:
            sublattices (list of Sublattice):
                list of active Sublattices to propose steps for. Active
                sublattices are those that include sites with configuration
                degrees of freedom DOFs, only occupancy on active sub-lattices'
                active sites are allowed to change.
            sublattice_probabilities (list of float): optional
                list of probabilities to pick a site from specific active
                sublattices.
            rng (np.Generator): optional
                The given PRNG must be the same instance as that used by the kernel and
                any bias terms, otherwise reproducibility will be compromised.
        """
        self.sublattices = sublattices
        self.active_sublattices = [
            sublatt for sublatt in self.sublattices if sublatt.is_active
        ]

        if sublattice_probabilities is None:
            self._sublatt_probs = np.array(
                len(self.active_sublattices)
                * [
                    1 / len(self.active_sublattices),
                ]
            )
        elif len(sublattice_probabilities) != len(self.active_sublattices):
            raise AttributeError(
                "Sublattice probabilities needs to be the same length as sublattices."
            )
        elif sum(sublattice_probabilities) != 1:
            raise ValueError("Sublattice probabilities must sum to one.")
        else:
            self._sublatt_probs = sublattice_probabilities

        self._rng = np.random.default_rng(rng)
        self.spec = Metadata(
            self.__class__.__name__,
            sublattices=[sublatt.species for sublatt in self.sublattices],
            sublattice_probabilities=self._sublatt_probs,
        )

    @property
    def sublattice_probabilities(self):
        """Get the sublattice probabilities."""
        return self._sublatt_probs

    @sublattice_probabilities.setter
    def sublattice_probabilities(self, value):
        """Set the sublattice probabilities."""
        if len(value) != len(self.active_sublattices):
            raise AttributeError(
                f"Can not set sublattice probabilities.\n Length must be the"
                f" same as the number of sublattices {len(self.sublattices)}"
            )
        if sum(value) != 1:
            raise ValueError(
                "Can not set sublattice probabilities.\n"
                "Sublattice probabilities must sum to one."
            )
        self._sublatt_probs = value

    @abstractmethod
    def propose_step(self, occupancy):
        """Propose an MCMC step.

        A step is given as a sequence of tuples, where each tuple is of the
        form (site index, species code to set)

        Args:
            occupancy (ndarray):
                encoded occupancy string.

        Returns:
            list(tuple): tuple of tuples each with (index, code)
        """
        return []

    def compute_log_priori_factor(self, occupancy, step):
        """Compute a-priori weight to adjust step acceptance ratio.

        This is essential in keeping detailed balance for some particular
        metropolis proposal methods. Return log of the factor for numerical
        accuracy.

        Args:
            occupancy (ndarray):
                encoded occupancy string
            step (list[tuple]):
                Metropolis step. list of tuples each with (index, code).

        Returns:
            float: log of a-priori adjustment weight.
        """
        return 0.0

    def update_aux_state(self, step, *args, **kwargs):
        """Update any auxiliary state information based on an accepted step."""
        return

    def set_aux_state(self, occupancy, *args, **kwargs):
        """Set the auxiliary occupancies from a checkpoint values."""
        return

    def get_random_sublattice(self):
        """Return a random sublattice based on given probabilities."""
        return self._rng.choice(self.active_sublattices, p=self._sublatt_probs)


class Flip(MCUsher):
    """Implementation of a simple flip step at a random site."""

    def propose_step(self, occupancy):
        """Propose a single random flip step.

        A step is given as a sequence of tuples, where each tuple is of the
        form (site index, species code to set)

        Args:
            occupancy (ndarray):
                encoded occupancy string.

        Returns:
            list(tuple): list of tuples each with (index, code)
        """
        sublattice = self.get_random_sublattice()
        site = self._rng.choice(sublattice.active_sites)
        choices = set(sublattice.encoding) - {occupancy[site]}
        return [(site, self._rng.choice(list(choices)))]


class Swap(MCUsher):
    """Implementation of a simple swap step for two random sites."""

    def propose_step(self, occupancy):
        """Propose a single random swap step.

        A step is given as a sequence of tuples, where each tuple is of the
        form (site index, species code to set)

        Args:
            occupancy (ndarray):
                encoded occupancy string.

        Returns:
            list(tuple): list of tuples each with (index, code)
        """
        sublattice = self.get_random_sublattice()
        site1 = self._rng.choice(sublattice.active_sites)
        species1 = occupancy[site1]
        sublattice_occu = occupancy[sublattice.active_sites]
        swap_options = sublattice.active_sites[sublattice_occu != species1]
        if swap_options.size > 0:  # check if swap_options are not empty
            site2 = self._rng.choice(swap_options)
            swap = [(site1, occupancy[site2]), (site2, species1)]
        else:
            # inefficient, maybe re-call method? infinite recursion problem
            swap = []
        return swap


class MultiStep(MCUsher):
    """A multistep usher to generate steps of several flips.

    Any usher can be used to chain steps, ie create multi flip steps, multi swap steps
    even multi composite steps, etc.
    """

    def __init__(
        self,
        sublattices,
        mcusher,
        step_lengths,
        step_probabilities=None,
        rng=None,
    ):
        """Initialize a multistep usher.

        Args:
            sublattices (list of Sublattice):
                list of active Sublattices to propose steps for. Active
                sublattices are those that include sites with configuration
                degrees of freedom DOFs, only occupancy on active sub-lattices' active sites
                are allowed to change.
            mcusher (MCUsher):
                An instantiated MCUsher to use to create multisteps.
            step_lengths (int or Sequence of int):
                The length or lengths of steps. If given a sequence of sizes, a size
                is chosen at each proposal according to step_probabilities. The step
                length proposed may be less in cases where the same site is chosen
                more than once. This should become less and less likely for larger
                supercells.
            step_probabilities (Sequence of float): optional
                A sequence of probabilities corresponding to the step lengths provided.
                If none is given, then a uniform probability is used.
            rng (np.Generator): optional
                The given PRNG must be the same instance as that used by the kernel and
                any bias terms, otherwise reproducibility will be compromised.
        """
        super().__init__(sublattices, rng=rng)

        self._sublatt_probs = None

        if isinstance(step_lengths, int):
            self._step_lens = np.array([step_lengths], dtype=int)
        else:
            self._step_lens = np.array(step_lengths, dtype=int)

        if step_probabilities is not None:
            if sum(step_probabilities) != 1.0:
                raise ValueError("The step_probabilities do not sum to 1.")
            if len(step_probabilities) != len(step_lengths):
                raise ValueError(
                    "The length of step_lengths and step_probabilities does not match."
                )
            self.step_p = np.array(step_probabilities)
        else:
            self._step_p = np.array(
                [1.0 / len(self._step_lens) for _ in range(len(self._step_lens))]
            )

        if isinstance(mcusher, str):  # TODO additional kwargs?
            mcusher = mcusher_factory(
                mcusher,
                self.sublattices,
            )

        self._mcusher = mcusher

        # update spec
        self.spec.step = self._mcusher.spec
        self.spec.step_lengths = self._step_lens
        self.spec.step_probabilities = self._step_p

    @property
    def sublattice_probabilities(self):
        """Get the probabilities of choosing a sublattice."""
        return self._mcusher.sublattice_probabilities

    @sublattice_probabilities.setter
    def sublattice_probabilities(self, value):
        """Set the probabilities of choosing a sublattice."""
        self._mcusher.sublattice_probabilities = value

    def propose_step(self, occupancy):
        """Propose a step given an occupancy."""
        step_length = self._rng.choice(self._step_lens, p=self._step_p)
        occu = occupancy.copy()

        steps = [self._mcusher.propose_step(occu)]
        for f in steps[-1]:
            occu[f[0]] = f[1]

        for _ in range(step_length - 1):
            step = self._mcusher.propose_step(occu)
            # only if all sites to be flipped are not already in steps
            if all(s not in (s for st in steps for s, _ in st) for s, _ in step):
                steps.append(step)
                for f in steps[-1]:
                    occu[f[0]] = f[1]

        # unpack them to single list
        return [flip for step in steps for flip in step]


class Composite(MCUsher):
    """A composite usher for chaining different step types.

    This can be used to mix step types for all sublattices or to create hybrid
    ensembles where different sublattices have different allowed steps.
    """

    def __init__(
        self,
        sublattices,
        mcushers,
        mcusher_weights=None,
        rng=None,
    ):
        """Initialize the composite usher.

        Args:
            sublattices (list of Sublattice):
                list of active Sublattices to propose steps for. Active
                sublattices are those that include sites with configuration
                degrees of freedom DOFs, only occupancy on active sub-lattices' active sites
                are allowed to change.
            mcushers (list of MCUsher):
                A list of mcushers to add to the composite.
            mcusher_weights (list of float):
                A list of the weights associated with each mcusher passed. With this
                the corresponding probabilities for each mcusher to be picked are
                generated. Must be in the same order as the mcusher list.
            rng (np.Generator): optional
                The given PRNG must be the same instance as that used by the kernel and
                any bias terms, otherwise reproducibility will be compromised.
        """
        super().__init__(sublattices, rng=rng)
        self._mcushers = []
        self._weights = []
        self._p = []
        self.spec.steps = []

        if mcusher_weights is None:
            mcusher_weights = len(mcushers) * [
                1,
            ]

        for weight, usher in zip(mcusher_weights, mcushers):
            if isinstance(usher, str):
                usher = mcusher_factory(
                    usher,
                    self.sublattices,
                )
            self.add_mcusher(usher, weight)
            self.spec.steps.append(usher.spec)

        # update spec
        self.spec.weights = self._weights

    @property
    def mcushers(self):
        """Get the list of mcushers."""
        return self._mcushers

    @property
    def weight(self):
        """Get the weights associated with each mcusher."""
        return self._weights

    def add_mcusher(self, mcusher, weight=1):
        """Add an MCUsher to the Composite usher.

        Args:
            mcusher (MCUsher):
                The MCUsher to add
            weight (float):
                The weight associated with the mcusher being added
        """
        self._mcushers.append(mcusher)
        self.spec.steps.append(mcusher.spec)
        self._update_p(weight)

    def _update_p(self, weight):
        """Update the probabilities for each mcuhser based on a new weight."""
        self._weights.append(weight)
        self.spec.weights = self._weights
        total = sum(self._weights)
        self._p = [weight / total for weight in self._weights]

    def propose_step(self, occupancy):
        """Propose a step given an occupancy."""
        return self._rng.choice(self._mcushers, p=self._p).propose_step(occupancy)


class TableFlip(MCUsher):
    """Implementation of table flips.

    Use user-specified or auto-calculated flip tables. Does not
    check the correctness of the flip tables.

    No longer computing and assigning a-priori proabilities
    before flip selection, and selects sites instead.

    Direct use of this class is not recommended, because the
    dict form of the flip table is not very easy to read and
    write.

    .. note::

    """

    def __init__(
        self,
        sublattices,
        rng=None,
        flip_table=None,
        charge_balanced=True,
        other_constraints=None,
        optimize_basis=False,
        table_ergodic=False,
        flip_weights=None,
        swap_weight=0.1,
    ):
        """Initialize TableFlip.

        Args:
            sublattices (list of Sublattice):
                list of Sublattices to propose steps for. Must
                be all sublattices, including active and inactive
                ones.
            rng (np.Generator): optional
                The given PRNG must be the same instance as that used by the
                kernel and any bias terms, otherwise reproducibility will be
                compromised.
            flip_table(2D np.ndarray[int]), optional:
                Flip vectors in "n" format, as generated by a CompositionSpace.
                If not given, we will initialize a CompositionSpace from
                sublattice information, and calculate the flip table
                with that CompositionSpace.
                Notice:
                1, We suggest you provide a pre-computed flip table, unless
                the number of sites in sub-lattices can be reduced to very
                small when divided with their GCD.
                2, As implemented in the Sublattice class,
                you are allowed to have multiple species on an inactive
                sub-lattice, but it is not the suggested approach.
                In this case, you have 3 choices. Either always try to split
                until all inactive sub-lattices only have one species
                (suggested), fixing number of inactive species as a
                constraint, or pre-compute your own flip table.
            charge_balanced(bool): optional
                Whether to add charge balance constraint. Default
                to true.
            other_constraints(List[tuple(1D arrayLike[int], int)]): optional
                Other integer constraints except charge balance and
                site-number conservation. Should be given in the form of
                tuple(a, bb), each gives constraint np.dot(a, n)=bb.
            optimize_basis(bool): optional
                Whether to optimize the basis to minimal flip sizes and maximal
                connectivity in the minimum super-cell size.
                When the minimal super-cell size is large, we recommend not to
                optimize basis.
            table_ergodic(bool): optional
                When generating a flip table, whether to add vectors and
                ensure ergodicity under a minimal super-cell size.
                Default to False.
                When the minimal super-cell size is large, we recommend not to
                ensure ergodicity. This is not only because of the computation
                difficulty; but also because at large super-cell size,
                the fraction of inaccessible compositions usually becomes
                minimal.
            swap_weight(float): optional
                Percentage of canonical swaps to attempt. Should be a
                positive value, but smaller than 1.
                Default to 0.1.
            flip_weights(1D Arraylike): optional
                Weights to adjust probability of each flip. If
                None given, will assign equal weights to each
                flip vector and its inverse. Length must equal to
                2*len(flip_table).

        .. note::
            In table flip, adding proposal bias to different
            sub-lattices is not meaningful. We will always assign
            sub-lattice probabilities as equal.
        """
        super().__init__(sublattices, rng=rng)
        self.bits = [sl.species for sl in self.sublattices]
        self.dim_ids = get_dim_ids_by_sublattice(self.bits)
        self.sublattice_sizes = np.array(
            [len(sl.sites) for sl in self.sublattices], dtype=int
        )
        self.supercell_size = np.gcd.reduce(self.sublattice_sizes)
        self.sublattice_sizes = self.sublattice_sizes // self.supercell_size
        self.max_n = [
            len(sublatt.active_sites)
            for sublatt in self.sublattices
            for _ in sublatt.species
        ]
        self.d = len(self.max_n)

        self._comp_space = CompositionSpace(
            self.bits,
            self.sublattice_sizes,
            charge_neutral=charge_balanced,
            other_constraints=other_constraints,
            optimize_basis=optimize_basis,
            table_ergodic=table_ergodic,
        )

        if flip_table is not None:
            self.flip_table = np.array(flip_table, dtype=int)
        else:
            self.flip_table = self._comp_space.flip_table

        self.swap_weight = swap_weight

        # A flip has two directions.
        if flip_weights is None:
            self.flip_weights = np.ones(len(self.flip_table) * 2)
        else:
            if (
                len(flip_weights) != len(self.flip_table)
                and len(flip_weights) != len(self.flip_table) * 2
            ):
                raise ValueError(
                    f"{len(flip_weights)} weights provided. "
                    "You must provide either 1* or 2* weights "
                    f"given {len(self.flip_table)} flip "
                    "vectors!"
                )
            # Forward and backward directions are assigned equal weights.
            if len(flip_weights) == len(self.flip_table):
                self.flip_weights = np.repeat(flip_weights, 2)
            if len(flip_weights) == 2 * len(self.flip_table):
                self.flip_weights = np.array(flip_weights)

        self._swapper = Swap(self.sublattices)
        self._dim_ids_table = get_dim_ids_table(self.sublattices, active_only=True)
        self._dim_ids_full = get_dim_ids_table(self.sublattices, active_only=False)

        # record specifications
        self.spec.flip_table = self.flip_table.tolist()
        self.spec.flip_weights = self.flip_weights.tolist()
        self.spec.other_constraints = jsanitize(other_constraints)
        self.spec.charge_balanced = charge_balanced
        self.spec.optimize_basis = optimize_basis
        self.spec.table_ergodic = table_ergodic
        self.spec.swap_weight = swap_weight

    def propose_step(self, occupancy):
        """Propose a single random flip step.

        A step is given as a sequence of tuples, where each tuple is of the
        form (site index, species code to set)

        1, Pick a flip in table;
        2, Choose sites from site specie statistics list according to the flip.
        3, (Acceptance proability will be adjusted according to a-priori
            weight to enforce detailed balance.)
        4, If selection failed, may or may not attempt a canonical swap based
           on self.add_swap parameter.

        There are circumstances where the current occupancy can not be accessed
        by minimal table flips. This method does not take care of that, please
        be careful when choosing starting compositions.

        Args:
            occupancy (ndarray):
                encoded occupancy string.

        Returns:
             list(tuple): list of tuples each with (idex, code).
        """
        rng = self._rng
        if rng.random() < self.swap_weight:
            return self._swapper.propose_step(occupancy)

        # We shall only flip active sites. And only active sites stated.
        species_list = occu_to_species_list(occupancy, self.d, self._dim_ids_table)
        species_n = [len(sites) for sites in species_list]
        species_list_full = occu_to_species_list(occupancy, self.d, self._dim_ids_full)
        species_n_full = [len(sites) for sites in species_list_full]

        if not np.allclose(
            self._comp_space._A @ np.array(species_n_full),
            self._comp_space._b * self.supercell_size,
        ):
            warnings.warn(
                "Current occupancy violates CompositionSpace constraints! "
                "Are you initializing trace?"
            )
            mask = np.zeros(2 * len(self.flip_table), dtype=int)
        else:
            mask = flip_weights_mask(self.flip_table, species_n, self.max_n).astype(int)
        masked_weights = self.flip_weights * mask
        # Mask out impossible selections.
        if np.any(masked_weights <= -NUM_TOL):
            raise ValueError(
                f"Masked weights: {masked_weights}" f" cannot have negative value!"
            )
        if np.allclose(masked_weights, 0):
            # Second condition to mute mckernel trace init.
            if not np.allclose(occupancy, 0):
                warnings.warn(
                    "Current occupancy is not ergodic! " "Will do canonical swap only!"
                )
            return self._swapper.propose_step(occupancy)

        idx = choose_section_from_partition(masked_weights, rng=rng)
        u = self.flip_table[idx // 2]
        # Forward or backward
        if idx % 2 == 1:
            u = -1 * u

        # Sub-lattices can not cross.
        step = []
        for sl_id, (sublatt, dim_ids) in enumerate(zip(self.sublattices, self.dim_ids)):
            if not sublatt.is_active:
                continue
            site_ids = []
            dim_ids = np.array(dim_ids, dtype=int)
            u_sl = u[dim_ids]
            dims_from = dim_ids[u_sl < 0]
            dims_to = dim_ids[u_sl > 0]
            codes_to = sublatt.encoding[u_sl > 0]
            for d in dims_from:
                site_ids.extend(
                    rng.choice(species_list[d], size=-1 * u[d], replace=False).tolist()
                )
            for d, code in zip(dims_to, codes_to):
                for site_id in rng.choice(site_ids, size=u[d], replace=False):
                    step.append((int(site_id), int(code)))
                    site_ids.remove(site_id)
            assert len(site_ids) == 0  # Site num conservation.

        return step

    def _get_flip_id(self, occupancy, step):
        """Compute flip id in table from occupancy and step."""
        dn = delta_counts_from_step(occupancy, step, self.d, self._dim_ids_table)

        if np.allclose(dn, 0):
            return -1, 0

        for fid, v in enumerate(self.flip_table):
            if np.allclose(v, dn):
                return fid, 0
            if np.allclose(-v, dn):
                return fid, 1

        return None, None

    def compute_log_priori_factor(self, occupancy, step):
        """Compute a-priori weight to adjust step acceptance ratio.

        This is essential in keeping detailed balance. Return log for
        numerical accuracy.

        Arg:
            occupancy (ndarray):
                encoded occupancy string. (Processor encoding.)
            step (list[tuple]):
                Metropolis step. list of tuples each with (index, code).
                (Processor encoding.)

        Returns:
            float: log of a-priori adjustment ratio.
        """
        fid, direction = self._get_flip_id(occupancy, step)
        if fid is None:
            raise ValueError(f"Step {step} is not in flip table.")

        if fid < 0:
            # Canonical swap, priori factor is always 1, log is 0.
            return 0

        u = (-2 * direction + 1) * self.flip_table[fid]

        n_now = occu_to_counts(occupancy, self.d, self._dim_ids_table)
        mask_now = flip_weights_mask(self.flip_table, n_now, self.max_n).astype(int)
        weights_now = self.flip_weights * mask_now
        p_now = (
            (1 - self.swap_weight)
            * weights_now[fid * 2 + direction]
            / weights_now.sum()
        )

        n_next = n_now + u
        mask_next = flip_weights_mask(self.flip_table, n_next, self.max_n).astype(int)
        weights_next = self.flip_weights * mask_next
        p_next = (
            (1 - self.swap_weight)
            * weights_next[fid * 2 + (1 - direction)]
            / weights_next.sum()
        )

        # Combinatorial factor.
        log_factor = np.log(p_next / p_now)
        dim_ids = np.arange(len(u), dtype=int)
        dims_nonzero = dim_ids[~np.isclose(u, 0)]

        def facln(n):  # log(n!), will speed up calculations.
            return gammaln(n + 1)

        for dim in dims_nonzero:
            log_factor += facln(n_now[dim]) - facln(n_next[dim])

        return log_factor


def mcusher_factory(usher_type, sublattices, *args, **kwargs):
    """Get a MC Usher from string name.

    Args:
        usher_type (str):
            string specifying step to instantiate.
        sublattices (list of Sublattice):
                list of Sublattices to propose steps for.
        *args:
            positional arguments passed to class constructor
        **kwargs:
            keyword arguments passed to class constructor

    Returns:
        MCUsher: instance of derived class.
    """
    usher_name = class_name_from_str(usher_type)
    return derived_class_factory(usher_name, MCUsher, sublattices, *args, **kwargs)


def available_step_types() -> tuple[str]:
    """Get available step types."""
    return get_subclasses_str(MCUsher)



"""Exotic cases where sub-lattices contains split and manually fixed sites."""

from itertools import product

import numpy as np
import pytest
from pymatgen.core import Lattice, Species, Structure

from smol.cofe import ClusterExpansion, ClusterSubspace
from smol.cofe.extern import EwaldTerm
from smol.cofe.space.domain import Vacancy
from smol.moca.ensemble import Ensemble

from .utils import get_random_solver_test_occu


# The following fixtures are designed specifically for testing the ground-state solver.
@pytest.fixture(scope="module")
def solver_test_prim():
    lat = Lattice.from_parameters(2, 2, 2, 60, 60, 60)
    return Structure(
        lat,
        [
            {
                "Li+": 1 / 5,
                "Mn2+": 1 / 5,
                "Mn4+": 1 / 5,
                "Ti4+": 1 / 5,
            },
            {
                "O2-": 1 / 2,
                "F-": 1 / 2,
            },
        ],
        [[0, 0, 0], [0.5, 0.5, 0.5]],
    )


@pytest.fixture(scope="module", params=["indicator", "sinusoid"])
def solver_test_subspace(solver_test_prim, request):
    # Use sinusoid basis to test if useful.
    space = ClusterSubspace.from_cutoffs(
        solver_test_prim, {2: 2.1, 3: 2.1}, basis=request.param
    )
    space.add_external_term(EwaldTerm())
    return space


@pytest.fixture(scope="module")
def solver_test_coefs(solver_test_subspace):
    solver_test_coefs = np.zeros(solver_test_subspace.num_corr_functions + 1)
    solver_test_coefs[0] = -10
    n_pair = len(solver_test_subspace.function_inds_by_size[2])
    n_tri = len(solver_test_subspace.function_inds_by_size[3])
    n_quad = 0
    i = 1
    solver_test_coefs[i : i + n_pair] = np.random.normal(size=n_pair)
    i += n_pair
    solver_test_coefs[i : i + n_tri] = np.random.normal(size=n_tri) * 0.4
    i += n_tri
    solver_test_coefs[i : i + n_quad] = np.random.normal(size=n_quad) * 0.1

    solver_test_coefs[-1] = 0.2
    return solver_test_coefs


@pytest.fixture(scope="module")
def solver_test_expansion(solver_test_subspace, solver_test_coefs):
    return ClusterExpansion(solver_test_subspace, solver_test_coefs)


@pytest.fixture(
    scope="module",
    params=list(product(["canonical", "semigrand"], ["expansion", "decomposition"])),
)
def orig_ensemble(solver_test_expansion, request):
    if request.param[0] == "semigrand":
        chemical_potentials = {
            "Li+": np.random.normal(),
            "Mn2+": np.random.normal(),
            "Mn4+": np.random.normal(),
            "Ti4+": np.random.normal(),
            "Vacancy": np.random.normal(),
            "O2-": np.random.normal(),
            "F-": np.random.normal(),
        }
    else:
        chemical_potentials = None
    return Ensemble.from_cluster_expansion(
        solver_test_expansion,
        np.diag([3, 2, 2]),
        request.param[1],
        chemical_potentials=chemical_potentials,
    )


@pytest.fixture(scope="module")
# Cation sub-lattice sorted in front of anion sub-lattice.
def orig_sublattices(orig_ensemble):
    return orig_ensemble.sublattices


@pytest.fixture(scope="module")
def solver_test_initial_occupancy(orig_sublattices):
    return get_random_solver_test_occu(orig_sublattices)


@pytest.fixture(scope="module")
def solver_test_ensemble(orig_ensemble, solver_test_initial_occupancy):
    cation_id = None
    anion_id = None
    for sl_id in range(2):
        if Species("Li", 1) in orig_ensemble.sublattices[sl_id].species:
            cation_id = sl_id
        if Species("O", -2) in orig_ensemble.sublattices[sl_id].species:
            anion_id = sl_id

    # Split the cation sublattice.
    new_ensemble = Ensemble.from_dict(orig_ensemble.as_dict())

    # Manually restrict 1 random li sites, 2 Vacancy site.
    cation_sites = new_ensemble.sublattices[cation_id].sites
    li_code = new_ensemble.sublattices[cation_id].encoding[
        new_ensemble.sublattices[cation_id].species.index(Species("Li", 1))
    ]
    li_sites = new_ensemble.sublattices[cation_id].sites[
        np.where(solver_test_initial_occupancy[cation_sites] == li_code)[0]
    ]
    li_restricts = np.random.choice(li_sites, size=1, replace=False)
    new_ensemble.restrict_sites(li_restricts)
    va_code = new_ensemble.sublattices[cation_id].encoding[
        new_ensemble.sublattices[cation_id].species.index(Vacancy())
    ]
    va_sites = new_ensemble.sublattices[cation_id].sites[
        np.where(solver_test_initial_occupancy[cation_sites] == va_code)[0]
    ]
    va_restricts = np.random.choice(va_sites, size=2, replace=False)
    new_ensemble.restrict_sites(va_restricts)

    ca_partitions = [
        [Species("Li", 1), Vacancy()],
        [Species("Mn", 2), Species("Mn", 4), Species("Ti", 4)],
    ]
    an_partitions = [[Species("O", -2)], [Species("F", -1)]]
    new_ensemble.split_sublattice_by_species(
        cation_id, solver_test_initial_occupancy, ca_partitions
    )
    # Sub-lattices all updated after partition.
    anion_id = anion_id if cation_id > anion_id else anion_id + 1
    new_ensemble.split_sublattice_by_species(
        anion_id, solver_test_initial_occupancy, an_partitions
    )

    # Check if sites a correctly restricted.
    assert len(new_ensemble.sublattices) == 4
    for site in li_restricts:
        assert site in new_ensemble.restricted_sites
    for site in va_restricts:
        assert site in new_ensemble.restricted_sites

    return new_ensemble



from abc import ABC, abstractmethod

import pytest

from smol.cofe.space.basis import (
    BasisIterator,
    ChebyshevIterator,
    IndicatorIterator,
    LegendreIterator,
    NumpyPolyIterator,
    PolynomialIterator,
    SinusoidIterator,
)
from smol.moca.kernel import Metropolis, UniformlyRandom, WangLandau
from smol.moca.kernel.base import MCKernelInterface, ThermalKernelMixin
from smol.moca.kernel.bias import (
    FugacityBias,
    MCBias,
    SquareChargeBias,
    SquareHyperplaneBias,
)
from smol.moca.kernel.mcusher import (
    Composite,
    Flip,
    MCUsher,
    MultiStep,
    Swap,
    TableFlip,
)
from smol.utils.class_utils import (
    class_name_from_str,
    derived_class_factory,
    get_subclasses,
)


@pytest.mark.parametrize(
    "class_str",
    [
        "TheBestClass",
        "The-Best-Class",
        "The-Best-class",
        "The-best-Class",
        "the-Best-Class",
        "The-best-class",
        "the-Best-class",
        "the-best-Class",
        "the-best-class",
    ],
)
def test_class_name_from_str(class_str):
    assert class_name_from_str(class_str) == "TheBestClass"


def test_get_subclasses():
    # test a few dummy classes
    class DummyABC(ABC):
        @abstractmethod
        def do_stuff(self):
            pass

    class DummyDummyABC(DummyABC):
        pass

    class DummyParent(DummyABC):
        def do_stuff(self):
            print("I'm doing stuff!")

    class DummyChild(DummyParent):
        def do_stuff(self):
            print("I'm doing more stuff!")

    assert all(
        c in get_subclasses(DummyABC).values() for c in [DummyParent, DummyChild]
    )
    assert DummyABC not in get_subclasses(DummyABC).values()
    assert DummyDummyABC not in get_subclasses(DummyABC).values()

    # now test classes in smol
    assert all(
        c in get_subclasses(MCKernelInterface).values()
        for c in [UniformlyRandom, Metropolis, WangLandau]
    )
    assert Metropolis in get_subclasses(ThermalKernelMixin).values()

    assert all(
        c in get_subclasses(MCUsher).values()
        for c in [Swap, Flip, MultiStep, Composite, TableFlip]
    )

    assert all(
        c in get_subclasses(MCBias).values()
        for c in [FugacityBias, SquareChargeBias, SquareHyperplaneBias]
    )

    assert all(
        c in get_subclasses(BasisIterator).values()
        for c in [
            SinusoidIterator,
            IndicatorIterator,
            LegendreIterator,
            ChebyshevIterator,
            PolynomialIterator,
        ]
    )

    # assert an abstract derived class is not included
    assert NumpyPolyIterator not in get_subclasses(BasisIterator).values()

    assert all(
        c in get_subclasses(BasisIterator).values()
        for c in [LegendreIterator, ChebyshevIterator, PolynomialIterator]
    )


def test_derived_class_factory(single_canonical_ensemble):
    kernel = derived_class_factory(
        "Metropolis",
        MCKernelInterface,
        single_canonical_ensemble,
        "swap",
        temperature=500,
    )
    assert isinstance(kernel, Metropolis)
    assert isinstance(kernel, MCKernelInterface)

    with pytest.raises(NotImplementedError):
        derived_class_factory(
            "BeepBoop", MCKernelInterface, single_canonical_ensemble, "swap"
        )



"""A uniformly random kernel.

The UniformlyRandom kernel is equivalent to a Metropolis kernel with infinite temperature.
"""

__author__ = "Luis Barroso-Luque"


from math import log

import numpy as np

from smol.moca.kernel.base import ALL_BIAS, ALL_MCUSHERS, MCKernel


class UniformlyRandom(MCKernel):
    """A Kernel that accepts all proposed steps.

    This kernel samples the random limit distribution where all states have the
    same probability (corresponding to an infinite temperature). If a
    bias is added then the corresponding distribution will be biased
    accordingly.
    """

    valid_mcushers = ALL_MCUSHERS
    valid_bias = ALL_BIAS

    def _accept_step(self, occupancy, step):
        exponent = self.mcusher.compute_log_priori_factor(occupancy, step)
        if self._bias is not None:
            exponent += self.trace.delta_trace.bias

        self.trace.accepted = np.array(
            True if exponent >= 0 else exponent > log(self._rng.random())
        )
        return self.trace.accepted



"""Implementation of Base MCKernel and Mixin classes."""

__author__ = "Luis Barroso-Luque"

from abc import ABC, ABCMeta, abstractmethod

import numpy as np

from smol.constants import kB
from smol.moca.kernel.bias import MCBias, mcbias_factory
from smol.moca.kernel.mcusher import MCUsher, mcusher_factory
from smol.moca.metadata import Metadata
from smol.moca.trace import StepTrace, Trace
from smol.utils.class_utils import class_name_from_str, get_subclasses

ALL_MCUSHERS = list(get_subclasses(MCUsher).keys())
ALL_BIAS = list(get_subclasses(MCBias).keys())


class MCKernelInterface(metaclass=ABCMeta):
    """MCKernelInterface for defining Monte Carlo Kernels.

    This interface defines the bare minimum methods and properties that must be
    implemented in MCKernel.
    """

    @property
    @abstractmethod
    def trace(self) -> StepTrace:
        """Return the StepTrace object used to record the values for the proposed step.

        The trace most often holds the occupancy, an acceptance value, the change in
        features and change in generalized enthalpy from the previous step.

        Basically, any value that is needed to compute thermodynamic properties
        from a MC sample should be recorded here.
        """

    @property
    @abstractmethod
    def spec(self) -> Metadata:
        """Return the Metadata object recording sampling specifications.

        Specifications should record all the necessary information to be able to
        reproduce the same results.
        """

    @abstractmethod
    def single_step(self, occupancy) -> StepTrace:
        """Carry out a single MC step.

        It is recommended to use a breakup of single step into methods as is done in
        the MCKernel base class, by using StandardSingleStepMixin:
        1) _compute_step_trace(occupancy, step)
        2) _accept_step(occupancy, step)
        3) _do_post_step()

        The above is not a hard requirement, as long as the StepTrace recording
        necessary values is returned, and the occupancy is modified in place
        according to the step. (not sure if that is the best design decision, if
        we decide it is not then we need to adjust the run method in Sampler)
        """

    @abstractmethod
    def compute_initial_trace(self, occupancy) -> Trace:
        """Compute an initial Trace.

        This function myst compute an initial Trace recording all the necessary absolute
        values that would be recorded during sampling using single_step. That is, this
        should not include "delta" or changes in values, but the absolute values of
        properties corresponding to the given occupancy.
        """

    @abstractmethod
    def set_aux_state(self, occupancy, *args, **kwargs):
        """Set the auxiliary occupancies from initial or checkpoint values.

        This class should be used to setup any auxiliary values that are not recorded
        in the traced values.
        """


class StandardSingleStepMixin(ABC):
    """A mixin interface class that breaks up the single_step method of a Kernel.

    The breakup is done using the following methods:

    1) _compute_step_trace(occupancy, step)
    2) _accept_step(occupancy, step)
    3) _do_post_step()

    The above methods must be implemented in derived classes.
    """

    @abstractmethod
    def _compute_step_trace(self, occupancy, step) -> None:
        """Compute the trace for a single step.

        Every quantity that is required to accept/reject should be computed here.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step
        """

    @abstractmethod
    def _accept_step(self, occupancy, step) -> bool:
        """Accept/reject a given step.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step
        """

    @abstractmethod
    def _do_accept_step(self, occupancy, step) -> np.ndarray:
        """Populate trace and aux states for an accepted step.

        All updates that should be done if a step has been accepted should be done here.

        The updated occupancy is returned.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step

        Returns:
            ndarray: new occupancy
        """

    def _do_post_step(self) -> None:
        """Do any post step updates.

        Any updates that should be done after a step has been attempted regardless of
        acceptance should be done here.
        """
        return

    def single_step(self, occupancy):
        """Attempt an MCMC step.

        Returns the next state in the chain and if the attempted step was
        successful.

        Args:
            occupancy (ndarray):
                encoded occupancy.

        Returns:
            StepTrace:
                A trace for states, traced values, and differences in traced values
                for the MC step
        """
        step = self.mcusher.propose_step(occupancy)
        self._compute_step_trace(occupancy, step)
        if self._accept_step(occupancy, step):
            occupancy = self._do_accept_step(occupancy, step)
        self._trace.occupancy = occupancy
        self._do_post_step()
        return self.trace


class MCKernel(StandardSingleStepMixin, MCKernelInterface, ABC):
    """Abstract base class for transition kernels.

    A kernel is used to implement a specific MC algorithm used to sample
    the ensemble classes. For an illustrative example of how to derive from this
    and write a specific kernel see the Metropolis kernel.
    """

    # Lists of valid helper classes, set these in derived kernels
    valid_mcushers = None
    valid_bias = None

    def __init__(
        self,
        ensemble,
        step_type,
        *args,
        seed=None,
        bias_type=None,
        bias_kwargs=None,
        **kwargs,
    ):
        """Initialize MCKernel.

        Args:
            ensemble (Ensemble):
                an Ensemble instance to obtain the features and parameters
                used in computing log probabilities.
            step_type (str):
                string specifying the MCMC step type.
            seed (int): optional
                non-negative integer to seed the PRNG
            bias_type (str): optional
                name for bias type instance.
            bias_kwargs (dict): optional
                dictionary of keyword arguments to pass to the bias
                constructor.
            args:
                positional arguments to instantiate the MCUsher for the
                corresponding step size.
            kwargs:
                keyword arguments to instantiate the MCUsher for the
                corresponding step size.
        """
        self.natural_params = ensemble.natural_parameters
        self._seed = seed if seed is not None else np.random.SeedSequence().entropy
        self._rng = np.random.default_rng(self._seed)
        self._ensemble = ensemble
        self._trace = StepTrace(accepted=np.array(True))
        self._usher, self._bias = None, None

        mcusher_name = class_name_from_str(step_type)
        self.mcusher = mcusher_factory(
            mcusher_name, ensemble.sublattices, *args, rng=self._rng, **kwargs
        )

        self._spec = Metadata(
            self.__class__.__name__, seed=self._seed, step=self.mcusher.spec
        )

        if bias_type is not None:
            bias_name = class_name_from_str(bias_type)
            bias_kwargs = {} if bias_kwargs is None else bias_kwargs
            self._bias = mcbias_factory(
                bias_name, ensemble.sublattices, rng=self._rng, **bias_kwargs
            )
            self.spec.bias = self.bias.spec

        # run a initial step to populate trace values
        # Enforce int32.
        _ = self.single_step(np.zeros(ensemble.num_sites, dtype=np.int32))

    @property
    def trace(self):
        """Return step trace."""
        return self._trace

    @property
    def spec(self):
        """Return sampling specifications."""
        return self._spec

    @property
    def ensemble(self):
        """Return the ensemble instance."""
        return self._ensemble

    @property
    def mcusher(self):
        """Get the MCUsher."""
        return self._usher

    @mcusher.setter
    def mcusher(self, usher):
        """Set the MCUsher."""
        if usher.__class__.__name__ not in self.valid_mcushers:
            raise ValueError(f"{type(usher)} is not a valid MCUsher for this kernel.")
        self._usher = usher

    @property
    def seed(self):
        """Get seed for PRNG."""
        return self._seed

    @property
    def bias(self):
        """Get the bias."""
        return self._bias

    @bias.setter
    def bias(self, bias):
        """Set the bias."""
        if bias.__class__.__name__ not in self.valid_bias:
            raise ValueError(f"{type(bias)} is not a valid MCBias for this kernel.")
        if "bias" not in self.trace.delta_trace.names:
            self.trace.delta_trace.bias = np.zeros(1)
        self._bias = bias

    def set_aux_state(self, occupancy, *args, **kwargs):
        """Set the auxiliary occupancies from initial or checkpoint values."""
        self.mcusher.set_aux_state(occupancy, *args, **kwargs)

    def _compute_step_trace(self, occupancy, step):
        """Compute the trace for a single step.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step
        """
        self.trace.delta_trace.features = self.ensemble.compute_feature_vector_change(
            occupancy, step
        )
        self.trace.delta_trace.enthalpy = np.array(
            np.dot(self.natural_params, self.trace.delta_trace.features),
            dtype=np.float64,
        )
        if self.bias is not None:
            self.trace.delta_trace.bias = np.array(
                self.bias.compute_bias_change(occupancy, step),
                dtype=np.float64,
            )
        # TODO consider adding bias = 0 when no bias is set to simplify code in
        #  single_step of kernels. may be faster maybe not...

    @abstractmethod
    def _accept_step(self, occupancy, step):
        """Accept/reject a given step.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step
        """
        return self.trace.accepted

    def _do_accept_step(self, occupancy, step):
        """Populate trace and aux states for an accepted step.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step

        Returns:
            ndarray: new occupancy
        """
        for site, species in step:
            occupancy[site] = species

        self.mcusher.update_aux_state(step)
        return occupancy

    def compute_initial_trace(self, occupancy):
        """Compute initial values for sample trace given an occupancy.

        Args:
            occupancy (ndarray):
                Initial occupancy

        Returns:
            Trace
        """
        trace = Trace()
        trace.occupancy = occupancy
        trace.features = self.ensemble.compute_feature_vector(occupancy)
        # set scalar values into shape (1,) array for sampling consistency.
        trace.enthalpy = np.array(
            [np.dot(self.natural_params, trace.features)], dtype=np.float64
        )
        if self.bias is not None:
            trace.bias = np.array([self.bias.compute_bias(occupancy)], dtype=np.float64)
        trace.accepted = np.array([True], dtype=bool)
        return trace


class ThermalKernelMixin:
    """Mixin class for transition kernels with a set temperature.

    Basically all kernels should use this Mixin class with the exception of those
    for multicanonical sampling and related methods (i.e. see Wang-Landau)

    Never really found a final say regarding Mixins with __init__ and attributes....
    ...oh well, here's one...
    """

    _kB: float = kB  # Boltzmann constant in eV/K

    def __init__(self, temperature, *args, **kwargs):
        """Initialize ThermalKernel.

        Args:
            ensemble (Ensemble):
                an Ensemble instance to obtain the features and parameters
                used in computing log probabilities.
            step_type (str):
                string specifying the MCMC step type.
            temperature (float):
                temperature at which the MCMC sampling will be carried out.
            args:
                positional arguments to instantiate the MCUsher for the
                corresponding step size.
            kwargs:
                keyword arguments to instantiate the MCUsher for the
                corresponding step size.
        """
        self.beta = 1.0 / (self.kB * temperature)
        super().__init__(*args, **kwargs)
        self.temperature = temperature

    @property
    def kB(self):
        """Get the temperature scale factor."""
        return self._kB

    @kB.setter
    def kB(self, kB):
        """Set the temperature scale factor."""
        self._kB = kB
        self.beta = 1.0 / (self.kB * self.temperature)

    @property
    def temperature(self):
        """Get the temperature of kernel."""
        return self.trace.temperature

    @temperature.setter
    def temperature(self, temperature):
        """Set the temperature and beta accordingly."""
        self.trace.temperature = np.array(temperature, dtype=np.float64)
        self.beta = 1.0 / (self.kB * temperature)

    def compute_initial_trace(self, occupancy):
        """Compute initial values for sample trace given occupancy.

        Args:
            occupancy (ndarray):
                Initial occupancy

        Returns:
            Trace
        """
        trace = super().compute_initial_trace(occupancy)
        trace.temperature = np.array([self.trace.temperature], dtype=np.float64)
        return trace


class MulticellKernel(StandardSingleStepMixin, MCKernelInterface, ABC):
    """Abstract MulticellKernel class.

    Use this to create kernels that apart from proposing new occupancy states, jumping
    among different kernels is also allowed.

    This is useful for example when one wants to sample over different supercell shapes,
    i.e. for generating special quasi-random structures.

    The ensembles in all kernels must have supercells of the same size, only the shape
    can change. All ensembles must be created from the same Hamiltonian. And all kernels
    must be of the same kind.

    There is no requirement for the individual kernels to have the same step type, bias,
    or any other setup parameter, which allows very flexible sampling strategies.
    However, users are responsible for ensuring that the overall sampling strategy is
    correct! (i.e. satisfies balance or detailed balance).
    """

    def __init__(
        self,
        mckernels,
        kernel_probabilities=None,
        kernel_hop_periods=5,
        kernel_hop_probabilities=None,
        seed=None,
    ):
        """Initialize MCKernel.

        Args:
            mckernels (list of MCKernels):
                a list of MCKernels instances to obtain the features and parameters
                used in computing log probabilities.
            seed (int): optional
                non-negative integer to seed the PRNG
            kernel_probabilities (Sequence of floats): optional
                Probability for choosing each of the supplied kernels
            kernel_hop_periods (int or Sequence of ints): optional
                number of steps between kernel hop attempts.
            kernel_hop_probabilities (Sequence of floats): optional
                probabilities for choosing each of the specified hop periods.
        """
        if any(not isinstance(kernel, type(mckernels[0])) for kernel in mckernels):
            raise ValueError("All kernels must be of the same type.")

        if any(
            kernel.ensemble.num_sites != mckernels[0].ensemble.num_sites
            for kernel in mckernels
        ):
            raise ValueError("All ensembles must have the same number of sites.")

        if any(
            not np.allclose(kernel.natural_params, mckernels[0].natural_params)
            for kernel in mckernels
        ):
            raise ValueError("All ensembles must have the same natural parameters.")

        if kernel_probabilities is not None:
            if sum(kernel_probabilities) != 1.0:
                raise ValueError("The kernel_probabilities do not sum to 1.")
            if len(kernel_probabilities) != len(mckernels):
                raise ValueError(
                    "The length of kernel_probabilities must be equal to the number of"
                    " mckernels."
                )
            self.kernel_p = np.array(kernel_probabilities)
        else:
            self._kernel_p = np.array(
                [1.0 / len(mckernels) for _ in range(len(mckernels))]
            )

        if isinstance(kernel_hop_periods, int):
            self._hop_periods = np.array([kernel_hop_periods], dtype=int)
        else:
            self._hop_periods = np.array(kernel_hop_periods, dtype=int)

        if kernel_hop_probabilities is not None:
            if sum(kernel_hop_probabilities) != 1.0:
                raise ValueError("The kernel_hop_probabilities do not sum to 1.")
            if len(kernel_hop_probabilities) != len(kernel_hop_periods):
                raise ValueError(
                    "The length of kernel_hop_periods and kernel_hop_probabilities does"
                    "not match."
                )
            self._hop_p = np.array(kernel_hop_probabilities)
        else:
            self._hop_p = np.array(
                [1.0 / len(self._hop_periods) for _ in range(len(self._hop_periods))]
            )

        self._seed = seed if seed is not None else np.random.SeedSequence().entropy
        self._rng = np.random.default_rng(self._seed)
        self._kernels = mckernels
        self._current_hop_period = self._rng.choice(self._hop_periods, p=self._hop_p)
        self._kernel_hop_counter = 1
        self._current_kernel_index = np.array(0, dtype=int)
        # keep values of proposed features and occupancy for cell hops
        self._new_features = None
        self._features = np.zeros(
            (len(mckernels), len(self._kernels[0].natural_params))
        )
        self._spec = Metadata(
            self.__class__.__name__,
            seed=self._seed,
            kernel_probabilities=self._kernel_p,
            kernel_hop_periods=self._hop_periods,
            kernel_hop_probabilities=self._hop_p,
            mckernels=[kernel.spec for kernel in mckernels],
        )
        # set kernel indices
        for i, kernel in enumerate(self._kernels):
            kernel.trace.kernel_index = np.array(i, dtype=int)
        self._trace = self._kernels[self._current_kernel_index].trace
        self._kernel_hop_counter = 1  # reset the counter

    @property
    def trace(self):
        """Return step trace."""
        return self._trace

    @property
    def spec(self):
        """Return sampling specifications."""
        return self._spec

    @property
    def mckernels(self):
        """Get the MCKernels."""
        return self._kernels

    @property
    def current_kernel(self):
        """Get the current kernel."""
        return self._kernels[self.trace.kernel_index]

    @property
    def ensemble(self):
        """Return the ensemble instance."""
        return self.current_kernel.ensemble

    @property
    def mcusher(self):
        """Get the MCUsher."""
        return self.current_kernel.mcusher

    @property
    def seed(self):
        """Get seed for PRNG."""
        return self._seed

    @property
    def bias(self):
        """Get the bias."""
        return self.current_kernel.bias

    @bias.setter
    def bias(self, bias):
        """Set the bias."""
        raise RuntimeError(
            "Cannot set the bias for a MultiCellKernel.\n"
            "You must set it for an individual kernel."
        )

    def _compute_step_trace(self, occupancy, step):
        """Compute the step trace.

        Args:
            occupancy (np.ndarray): the current occupancy
            step (np.ndarray): the proposed step

        Returns:
            StepTrace: the step trace
        """
        occupancy = occupancy.copy()  # make a copy to avoid modifying the original
        for site, species in step:
            occupancy[site] = species

        self._new_features = self.ensemble.compute_feature_vector(occupancy)
        # from kernel currently at
        prev_features = self._features[self._current_kernel_index]
        self.trace.delta_trace.features = self._new_features - prev_features
        delta_enthalpy = np.dot(
            self.trace.delta_trace.features, self.ensemble.natural_parameters
        )
        self.trace.delta_trace.enthalpy = delta_enthalpy

    def _do_accept_step(self, occupancy, step):
        """Populate trace and aux states for an accepted step.

        Args:
            occupancy (ndarray):
                Current occupancy
            step (Sequence of tuple):
                Sequence of tuples of ints representing an MC step

        Returns:
            ndarray: new occupancy
        """
        for site, species in step:
            occupancy[site] = species

        # update the features
        self._features[self.trace.kernel_index, :] = self._new_features
        return occupancy

    def single_step(self, occupancy):
        """Attempt an MCMC step.

        Returns the next state in the chain and if the attempted step was
        successful.

        Args:
            occupancy (ndarray):
                encoded occupancy.

        Returns:
            StepTrace: a step trace for states and traced values for a single step
        """
        # check if we should attempt to hop kernels
        if self._kernel_hop_counter % self._current_hop_period == 0:
            # this is the trace for the new kernel (the one attempting to hop to)
            # in some cases it may hop to the same kernel, but that is ok
            self._trace = self._rng.choice(self._kernels, p=self._kernel_p).trace
            # set the occupancy to the one currently in the trace of the kernel
            # attempting to hop to, not the one passed from the previous step
            self._trace = super().single_step(self.trace.occupancy)
            if self.trace.accepted:
                # update the current kernel index
                self._current_kernel_index = self.trace.kernel_index
                # modify the occupancy in place, to follow MCKernelInterface
                occupancy[:] = self.trace.occupancy
            else:
                # reset the trace to the one of the previous kernel if step not accepted
                self._trace = self._kernels[self._current_kernel_index].trace
                self._trace.occupancy = occupancy
                self._trace.accepted = np.array(False)

            # reset the hop period and counter
            self._current_hop_period = self._rng.choice(
                self._hop_periods, p=self._hop_p
            )
            self._kernel_hop_counter = 1
        else:  # if not do a single step with current kernel
            self._trace = self.current_kernel.single_step(occupancy)
            self._kernel_hop_counter += 1
            # We need to save the latest property...
            if self.trace.accepted:
                self._features[
                    self._current_kernel_index
                ] += self.trace.delta_trace.features

        # assert self.trace.kernel_index == self._current_kernel_index
        return self.trace

    def set_aux_state(self, occupancy, *args, **kwargs):
        """Set the occupancies, features and enthalpy for each supercell kernel.

        Since multiple underlying kernels exist, occupancy can be a 2D array
        of occupancies for each kernel, ie dim = (n_kernels, n_sites).
        """
        # set the current and previous values based on given occupancy
        # Enforce int32.
        if occupancy.dtype != np.int32:
            occupancy = occupancy.astype(np.int32)
        new_features = []
        if occupancy.ndim == 2 and occupancy.shape[0] == len(self._kernels):
            for kernel, occupancy in zip(self._kernels, occupancy):
                # Enforce int32.
                occupancy = np.ascontiguousarray(occupancy, dtype=np.int32)
                kernel.trace.occupancy = occupancy
                kernel.set_aux_state(occupancy, *args, **kwargs)
                new_features.append(kernel.ensemble.compute_feature_vector(occupancy))
            self._features = np.vstack(new_features)
        else:  # if only one assume it is continuing from a run...
            self._features[self._current_kernel_index, :] = (
                self.current_kernel.ensemble.compute_feature_vector(occupancy)
            )
            self.current_kernel.set_aux_state(occupancy, *args, **kwargs)

    def compute_initial_trace(self, occupancy):
        """Compute the initial trace for a given occupancy."""
        trace = self.current_kernel.compute_initial_trace(occupancy)
        trace.kernel_index = self._current_kernel_index
        return trace



"""Handle conversion of variable indices."""

from typing import List, Tuple

import numpy as np
from pymatgen.core import Structure

from smol.cofe.space.domain import Vacancy, get_allowed_species
from smol.moca.ensemble import Sublattice

__author__ = "Fengyu Xie"


def get_variable_indices_for_each_composition_component(
    sublattices: List[Sublattice],
    variable_indices: List[List[int]],
    structure: Structure,
) -> List[Tuple[List[int], int]]:
    """Get variables and the number of restricted sites for each composition component.

    Composition components are the components in a vector representation of composition
    defined as the "counts" format in moca.composition.
    Args:
        sublattices(list of Sublattice):
            Sub-lattices to build the upper-bound problem on.
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
        structure(Structure):
            The supercell structure stored in a processor's structure attribute.
            The sub-lattices must match the processor structure, or they must be the result
            of splitting with the initial_occupancy. See smol.moca.sublattice for the
            explanation of splitting a sub-lattice.
    Returns:
        list of tuples of (list of int, int):
            A list of tuples containing indices of variables falling under each
            composition component, and the number of sites manually restricted
            or naturally inactive to always be occupied by the species corresponding
            to this component. Used to quickly generate composition constraints.
    """
    orig_site_spaces = get_allowed_species(structure)

    var_ids_for_dims = []
    for sublattice in sublattices:
        for species in sublattice.species:
            active_var_inds = []
            num_fixed = 0
            for site_id in sublattice.sites:
                orig_site_space = orig_site_spaces[site_id]
                orig_sp_id = orig_site_space.index(species)
                var_ind = variable_indices[site_id][orig_sp_id]
                # Active variable.
                if var_ind >= 0:
                    active_var_inds.append(var_ind)
                # Always true.
                elif var_ind == -1:
                    num_fixed += 1
            var_ids_for_dims.append((active_var_inds, num_fixed))

    return var_ids_for_dims


def map_ewald_indices_to_variable_indices(
    structure: Structure,
    variable_indices: List[List[int]],
) -> List[int]:
    """Map row indices in ewald matrix to indices of boolean variables.

    Args:
        structure(Structure):
            The structure attribute of ensemble Processor. This is required
            to correctly parse the rows in the ewald_structure of an
            EwaldTerm, as the given sub-lattices might come form a split
            ensemble, whose site_spaces can not match the ewald_structure.
        variable_indices(list of lists of int):
            List of variable indices corresponding to each active site index and
            index of species in its site space. Inactive sites will be marked by
            either -1 or -2. See documentation in groundstate.upper_bound.variables.
    Returns:
        list of int:
            Index of cvxpy variable corresponding to each ewald matrix row. If a
            row corresponds to a site with only one allowed species or the species
            on a manually restricted site in a pre-defined initial occupancy,
            it will be marked as -1 (always occupied).
            If a row corresponds to a site manually restricted to
            be occupied by other species rather than the species corresponding to
            this row, it will be marked as -2 (never occupied).
            This is to quickly parse the ewald term into the ewald objective
            function.
    """
    num_sites = len(variable_indices)

    orig_site_spaces = get_allowed_species(structure)

    ewald_ids_to_var_ids = []
    for site_id in range(num_sites):
        for orig_sp_id, orig_species in enumerate(orig_site_spaces[site_id]):
            # Vacancies are always skipped.
            if isinstance(orig_species, Vacancy):
                continue
            ewald_ids_to_var_ids.append(variable_indices[site_id][orig_sp_id])

    return ewald_ids_to_var_ids


def get_sublattice_indices_by_site(sublattices: List[Sublattice]) -> np.array:
    """Get the indices of sub-lattice to which each site belong.

    Args:
        sublattices(list of Sublattice):
            Sub-lattices to build the upper-bound problem on.
            Must contain all sites in the ensemble, and the site indices must start from
            0 and be continuous.
    Returns:
        1D np.ndarray of int:
            An array containing the sub-lattice index where each site belongs to.
            Used to quickly access the sub-lattice information.
    """
    num_sites = sum([len(sublattice.sites) for sublattice in sublattices])
    site_sublattice_ids = np.zeros(num_sites, dtype=int) - 1
    for sublattice_id, sublattice in enumerate(sublattices):
        site_sublattice_ids[sublattice.sites] = sublattice_id
    if np.any(site_sublattice_ids < 0):
        raise ValueError(
            f"Provided sub-lattices: {sublattices} do not contain all"
            f" sites in ensemble, or the site-indices are not continuous!"
        )
    return site_sublattice_ids



import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.analysis.structure_matcher import StructureMatcher

from smol.capp.generate.random import _gen_unconstrained_ordered_occu
from smol.cofe import ClusterExpansion
from smol.cofe.extern import EwaldTerm
from smol.cofe.space.domain import Vacancy, get_allowed_species
from smol.moca.processor import (
    ClusterDecompositionProcessor,
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
)
from smol.moca.processor.base import Processor
from smol.moca.processor.distance import (
    ClusterInteractionDistanceProcessor,
    CorrelationDistanceProcessor,
)
from smol.utils._openmp_helpers import _openmp_effective_numthreads
from smol.utils.cluster.numthreads import DEFAULT_NUM_THREADS
from tests.utils import assert_msonable, assert_pickles, gen_random_ordered_structure

pytestmark = pytest.mark.filterwarnings("ignore:All bit combos have been removed")

RTOL = 1e-12  # relative tolerance to check property change functions
# absolute tolerance to check property change functions (eps is approx 2E-16)
ATOL = 2e4 * np.finfo(float).eps
DRIFT_TOL = 10 * np.finfo(float).eps  # tolerance of average drift


@pytest.fixture
def ce_processor(cluster_subspace, rng):
    coefs = 2 * rng.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    return ClusterExpansionProcessor(
        cluster_subspace, supercell_matrix=scmatrix, coefficients=coefs
    )


@pytest.fixture(params=["real", "reciprocal", "point"])
def ewald_processor(cluster_subspace, rng, request):
    coef = rng.random(1)
    scmatrix = 3 * np.eye(3)
    ewald_term = EwaldTerm(use_term=request.param)
    return EwaldProcessor(
        cluster_subspace,
        supercell_matrix=scmatrix,
        coefficient=coef,
        ewald_term=ewald_term,
    )


@pytest.fixture(params=["correlation", "interaction"])
def processor_distance_processor(cluster_subspace, rng, request):
    # return a processor and the corresponding distance processor
    scmatrix = 3 * np.eye(3)
    coefs = np.ones(len(cluster_subspace))
    if request.param == "correlation":
        target_weights = rng.random(len(cluster_subspace) - 1)
        procs = (
            ClusterExpansionProcessor(cluster_subspace, scmatrix, coefs),
            CorrelationDistanceProcessor(
                cluster_subspace, scmatrix, target_weights=target_weights
            ),
        )
    else:
        expansion = ClusterExpansion(cluster_subspace, coefs)
        target_weights = rng.random(cluster_subspace.num_orbits - 1)
        procs = (
            ClusterDecompositionProcessor(
                cluster_subspace, scmatrix, expansion.cluster_interaction_tensors
            ),
            ClusterInteractionDistanceProcessor(
                cluster_subspace,
                scmatrix,
                expansion.cluster_interaction_tensors,
                target_weights=target_weights,
            ),
        )
    return procs


def test_encode_decode_property(ce_processor, rng):
    occu = _gen_unconstrained_ordered_occu(ce_processor.get_sublattices(), rng=rng)
    decoccu = ce_processor.decode_occupancy(occu)
    for species, space in zip(decoccu, ce_processor.allowed_species):
        assert species in space
    npt.assert_equal(occu, ce_processor.encode_occupancy(decoccu))


def test_site_spaces(ce_processor):
    assert all(
        sp in ce_processor.structure.composition
        for space in ce_processor.unique_site_spaces
        for sp in space
        if sp != Vacancy()
    )
    assert all(
        sp in ce_processor._subspace.expansion_structure.composition
        for space in ce_processor.active_site_spaces
        for sp in space
        if sp != Vacancy()
    )


def test_sublattice(ce_processor):
    sublattices = ce_processor.get_sublattices()
    # These are default initialized, not split.
    site_species = get_allowed_species(ce_processor.structure)
    for sublatt, site_space in zip(sublattices, ce_processor.unique_site_spaces):
        assert sublatt.site_space == site_space
        for site in sublatt.sites:
            assert site_species[site] == list(site_space.keys())
    assert sum(len(sublatt.sites) for sublatt in sublattices) == len(
        ce_processor.structure
    )


def test_structure_occupancy_conversion(ce_processor, rng):
    sm = StructureMatcher()
    for _ in range(10):
        s_init = gen_random_ordered_structure(
            ce_processor.cluster_subspace.structure,
            size=ce_processor.supercell_matrix,
            rng=rng,
        )
        s_init = s_init.get_sorted_structure()
        occu_init = ce_processor.occupancy_from_structure(s_init)

        s_conv = ce_processor.structure_from_occupancy(occu_init)
        s_conv = s_conv.get_sorted_structure()

        # occu_conv = ce_processor.occupancy_from_structure(s_conv)

        # For symmetrically equivalent structures, StructureMatcher might generate
        # different structure_site_mappings
        # (see cluster_subspace.structure_site_mappings), therefore we may get
        # different occupancy strings with occupancy_from_structure, and
        # occu1 -> structure -> occu2 conversion cycle does not guarantee that
        # occu1 == occu2. In most use cases, it is not necessary to enforce that
        # occu1 == occu2. If you have to do so, you'll need to deeply modify the code of
        # StructureMatcher, which might not be a trivial task. Here we will only test
        # whether occu1 -> str1 and occu2 -> str2 are symmetrically equivalent.
        # This should be enough in our application. We notify the users about this
        # mismatch in the documentations.
        assert sm.fit(s_init, s_conv)


# only test composite_processor predicting an energy value
def test_compute_energy(composite_processor, rng):
    occu = _gen_unconstrained_ordered_occu(
        composite_processor.get_sublattices(), rng=rng
    )
    struct = composite_processor.structure_from_occupancy(occu)
    pred = np.dot(
        composite_processor.raw_coefs,
        composite_processor.cluster_subspace.corr_from_structure(struct, False),
    )
    assert composite_processor.compute_property(occu) == pytest.approx(pred, abs=ATOL)


# General tests for all processors
# Currently being done only on composites because I can not for the life of
# me figure out a clean way to parametrize with parametrized fixtures or use a
# fixture union from pytest_cases that works.


def test_get_average_drift(processor):
    forward, reverse = processor.compute_average_drift()
    assert forward <= DRIFT_TOL and reverse <= DRIFT_TOL


def test_compute_property_change(processor, rng):
    sublattices = processor.get_sublattices()
    occu = _gen_unconstrained_ordered_occu(sublattices, rng=rng)
    occu_wrong_type = occu.astype(np.int64)
    active_sublattices = [sublatt for sublatt in sublattices if sublatt.is_active]

    for _ in range(100):
        sublatt = rng.choice(active_sublattices)
        site = rng.choice(sublatt.sites)
        new_sp = rng.choice(sublatt.encoding)
        new_occu = occu.copy()
        new_occu[site] = new_sp
        prop_f = processor.compute_property(new_occu)
        prop_i = processor.compute_property(occu)
        dprop = processor.compute_property_change(occu, [(site, new_sp)])
        # Check error raising behavior.
        if isinstance(
            processor, (ClusterExpansionProcessor, ClusterDecompositionProcessor)
        ):
            with pytest.raises(ValueError):
                _ = processor.compute_property(occu_wrong_type)
        if isinstance(
            processor,
            (ClusterExpansionProcessor, ClusterDecompositionProcessor, EwaldProcessor),
        ):
            with pytest.raises(ValueError):
                _ = processor.compute_property_change(occu_wrong_type, [(site, new_sp)])

        # Check with some tight tolerances.
        npt.assert_allclose(dprop, prop_f - prop_i, rtol=RTOL, atol=ATOL)
        # Test reverse matches forward
        old_sp = occu[site]
        rdprop = processor.compute_property_change(new_occu, [(site, old_sp)])
        npt.assert_allclose(dprop, -1 * rdprop, rtol=RTOL, atol=ATOL)


def test_compute_feature_change(processor, rng):
    sublattices = processor.get_sublattices()
    occu = _gen_unconstrained_ordered_occu(sublattices, rng=rng)
    active_sublattices = [sublatt for sublatt in sublattices if sublatt.is_active]
    processor.cluster_subspace.change_site_bases("indicator")

    for _ in range(100):
        sublatt = rng.choice(active_sublattices)
        site = rng.choice(sublatt.sites)
        new_sp = rng.choice(sublatt.encoding)
        new_occu = occu.copy()
        new_occu[site] = new_sp
        feat_f = processor.compute_feature_vector(new_occu)
        feat_i = processor.compute_feature_vector(occu)
        dfeat = processor.compute_feature_vector_change(occu, [(site, new_sp)])
        # Check with some tight tolerances.
        npt.assert_allclose(dfeat, feat_f - feat_i, rtol=RTOL, atol=ATOL)
        # Test reverse matches forward
        old_sp = occu[site]
        rdfeat = processor.compute_feature_vector_change(new_occu, [(site, old_sp)])
        npt.assert_allclose(dfeat, -1 * rdfeat, rtol=RTOL, atol=ATOL)


def test_msonable(processor, rng):
    occu = _gen_unconstrained_ordered_occu(processor.get_sublattices(), rng=rng)
    d = processor.as_dict()
    pr = Processor.from_dict(d)
    assert processor.compute_property(occu) == pr.compute_property(occu)
    npt.assert_array_equal(processor.coefs, pr.coefs)
    # send in pr bc composite_processor is scoped for function and new random
    # coefficients will be created.
    assert_msonable(pr)


def test_pickles(processor, rng):
    assert_pickles(processor)


# ClusterExpansionProcessor only tests
def test_compute_correlation_vector(ce_processor, rng):
    occu = _gen_unconstrained_ordered_occu(ce_processor.get_sublattices(), rng=rng)
    struct = ce_processor.structure_from_occupancy(occu)
    # same as normalize=False in corr_from_structure
    npt.assert_allclose(
        ce_processor.compute_feature_vector(occu) / ce_processor.size,
        ce_processor.cluster_subspace.corr_from_structure(struct),
    )

    npt.assert_allclose(
        ce_processor.compute_feature_vector(occu) / ce_processor.size,
        ce_processor.cluster_subspace.corr_from_structure(struct),
    )


# cluster decomposition processor
def test_compute_cluster_interactions(cluster_subspace, rng):
    coefs = 2 * np.random.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    expansion = ClusterExpansion(cluster_subspace, coefs)
    processor = ClusterDecompositionProcessor(
        cluster_subspace, scmatrix, expansion.cluster_interaction_tensors
    )

    occu = _gen_unconstrained_ordered_occu(processor.get_sublattices(), rng=rng)
    struct = processor.structure_from_occupancy(occu)
    # same as normalize=False in corr_from_structure
    proc_interactions = processor.compute_feature_vector(occu) / processor.size
    exp_interactions = expansion.cluster_interactions_from_structure(struct)
    npt.assert_allclose(proc_interactions, exp_interactions)
    pred_energy = expansion.predict(struct, normalized=True)
    assert sum(
        cluster_subspace.orbit_multiplicities * proc_interactions
    ) == pytest.approx(pred_energy, abs=ATOL)


def test_bad_coef_length(cluster_subspace, rng):
    coefs = rng.random(cluster_subspace.num_corr_functions - 1)
    with pytest.raises(ValueError):
        ClusterExpansionProcessor(cluster_subspace, 5 * np.eye(3), coefficients=coefs)


def test_bad_composite(cluster_subspace, rng):
    coefs = 2 * rng.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    proc = CompositeProcessor(cluster_subspace, supercell_matrix=scmatrix)
    with pytest.raises(AttributeError):
        proc.add_processor(
            CompositeProcessor(cluster_subspace, supercell_matrix=scmatrix)
        )
    with pytest.raises(ValueError):
        proc.add_processor(
            ClusterExpansionProcessor(
                cluster_subspace, 2 * scmatrix, coefficients=coefs
            )
        )
    with pytest.raises(ValueError):
        new_cs = cluster_subspace.copy()
        ids = range(1, new_cs.num_corr_functions)
        new_cs.remove_corr_functions(rng.choice(ids, size=10))
        proc.add_processor(
            ClusterExpansionProcessor(new_cs, scmatrix, coefficients=coefs)
        )


def test_distance_processor(processor_distance_processor, rng):
    # test distance processor results vs compute directly from the corresponding
    # processor
    processor, distance_processor = processor_distance_processor
    for _ in range(5):
        occu = _gen_unconstrained_ordered_occu(processor.get_sublattices(), rng=rng)

        # remove first entry since it is the "exact match diameter" in the distance metric
        expected = abs(
            processor.compute_feature_vector(occu)[1:] / processor.size
            - distance_processor.target_vector[1:]
        )
        actual = distance_processor.compute_feature_vector(occu)
        npt.assert_allclose(actual[1:], expected)

        diameter = distance_processor.exact_match_max_diameter(actual)
        expected = np.dot(
            distance_processor.coefs, np.concatenate([[diameter], expected])
        )
        actual = distance_processor.compute_property(occu)
        assert actual == pytest.approx(expected)

        for _ in range(100):
            active_sublattices = [
                sublatt for sublatt in processor.get_sublattices() if sublatt.is_active
            ]
            sublatt = rng.choice(active_sublattices)
            site = rng.choice(sublatt.sites)
            new_sp = rng.choice(sublatt.encoding)
            new_occu = occu.copy()
            new_occu[site] = new_sp
            flips = [(site, new_sp)]

            # remove first element since it is the "exact match diameter" in the distance metric
            dist_f = abs(
                processor.compute_feature_vector(new_occu)[1:] / processor.size
                - distance_processor.target_vector[1:]
            )
            dist_i = abs(
                processor.compute_feature_vector(occu)[1:] / processor.size
                - distance_processor.target_vector[1:]
            )

            distances = distance_processor.compute_feature_vector_distances(occu, flips)
            npt.assert_allclose(distances[0][1:], dist_i, rtol=RTOL, atol=ATOL)
            npt.assert_allclose(distances[1][1:], dist_f, rtol=RTOL, atol=ATOL)

            diameter_i = distance_processor.exact_match_max_diameter(distances[0])
            diameter_f = distance_processor.exact_match_max_diameter(distances[1])

            expected = np.dot(
                distance_processor.coefs,
                np.concatenate([[diameter_f - diameter_i], dist_f - dist_i]),
            )
            actual = distance_processor.compute_property_change(occu, flips)
            assert actual == pytest.approx(expected)


def test_exact_match_max_diameter(processor_distance_processor, rng):
    processor, distance_processor = processor_distance_processor

    # check a vector with all zeros returns the max diameter
    distance_vector = np.zeros(len(processor.coefs))
    max_diameter = max(processor.cluster_subspace.orbits_by_diameter.keys())
    assert distance_processor.exact_match_max_diameter(distance_vector) == max_diameter

    # check a random orbit in between
    # exclude zero diameter, and smallest diameter orbit
    diameter = rng.choice(
        list(processor.cluster_subspace.orbits_by_diameter.keys())[2:]
    )
    orbit = rng.choice(processor.cluster_subspace.orbits_by_diameter[diameter])

    if isinstance(distance_processor, CorrelationDistanceProcessor):
        # this is correlation based
        index = rng.choice(range(orbit.bit_id, orbit.bit_id + len(orbit)))
    else:
        index = orbit.id

    distance_vector[index] = 2 * distance_processor.match_tol
    assert 0 < distance_processor.exact_match_max_diameter(distance_vector) < diameter

    # check a vector exceeding match_tol for first point orbit
    distance_vector[1] = 2 * distance_processor.match_tol
    assert distance_processor.exact_match_max_diameter(distance_vector) == 0.0


def test_bad_distance_processor(single_subspace, rng):
    scm = 3 * np.eye(3)
    subspace = single_subspace.copy()

    with pytest.raises(ValueError):
        subspace.add_external_term(EwaldTerm())
        proc = CorrelationDistanceProcessor(subspace, scm)

    with pytest.raises(ValueError):
        proc = CorrelationDistanceProcessor(single_subspace, scm, match_weight=-1)

    with pytest.raises(ValueError):
        target_weights = np.ones(len(single_subspace) - 4)
        proc = CorrelationDistanceProcessor(
            single_subspace, scm, target_weights=target_weights
        )


def test_set_threads(single_subspace):
    coefs = 2 * np.random.random(single_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    expansion = ClusterExpansion(single_subspace, coefs)

    ceproc = ClusterExpansionProcessor(single_subspace, scmatrix, coefs)
    cdproc = ClusterDecompositionProcessor(
        single_subspace,
        scmatrix,
        expansion.cluster_interaction_tensors,
    )

    # assert defaults
    assert ceproc.num_threads == DEFAULT_NUM_THREADS
    assert ceproc._evaluator.num_threads == ceproc.num_threads
    assert ceproc.num_threads_full == DEFAULT_NUM_THREADS

    for eval_data in ceproc._eval_data_by_sites.values():
        assert eval_data.evaluator.num_threads == ceproc.num_threads

    assert cdproc.num_threads == DEFAULT_NUM_THREADS
    assert ceproc.num_threads_full == DEFAULT_NUM_THREADS

    for eval_data in cdproc._eval_data_by_sites.values():
        assert eval_data.evaluator.num_threads == cdproc.num_threads

    # assert setting thread values
    ceproc.num_threads = 1
    cdproc.num_threads = 1

    ceproc.num_threads_full = 1
    cdproc.num_threads_full = 1

    assert ceproc.num_threads == 1
    assert ceproc.num_threads_full == 1

    for eval_data in ceproc._eval_data_by_sites.values():
        assert eval_data.evaluator.num_threads == 1

    assert cdproc.num_threads == 1
    assert ceproc.num_threads_full == 1

    for eval_data in ceproc._eval_data_by_sites.values():
        assert eval_data.evaluator.num_threads == 1

    # assert -1 gives max number of threads
    ceproc.num_threads = -1
    assert ceproc.num_threads == _openmp_effective_numthreads(n_threads=-1)

    # assert setting more complains
    with pytest.warns(UserWarning):
        ceproc.num_threads = _openmp_effective_numthreads() + 1
        assert ceproc.num_threads == _openmp_effective_numthreads()



"""Test smol.moca.utils.occu_utils."""

import itertools

import numpy as np
import numpy.testing as npt
import pytest

from smol.capp.generate.random import _gen_unconstrained_ordered_occu
from smol.moca.occu_utils import (
    delta_counts_from_step,
    get_dim_ids_by_sublattice,
    get_dim_ids_table,
    occu_to_counts,
    occu_to_species_list,
)


@pytest.fixture(scope="module")
def sublattices(ensemble):
    sublattices = ensemble.sublattices
    all_sites = [site for sublatt in sublattices for site in sublatt.sites]
    sites_to_restrict = np.random.choice(
        all_sites, size=len(all_sites) // 2, replace=False
    )
    ensemble.restrict_sites(sites_to_restrict)
    return ensemble.sublattices


@pytest.fixture(scope="module")
def bits(sublattices):
    return [sl.species for sl in sublattices]


@pytest.fixture(scope="module")
def dim_ids(bits):
    return get_dim_ids_by_sublattice(bits)


@pytest.fixture(scope="module")
def dim_ids_table_active(sublattices):
    return get_dim_ids_table(sublattices, active_only=True)


@pytest.fixture(scope="module")
def dim_ids_table_all(sublattices):
    return get_dim_ids_table(sublattices, active_only=False)


def test_dim_ids(dim_ids, bits):
    d = sum(len(species) for species in dim_ids)
    for sl_id, dim_id in enumerate(dim_ids):
        assert len(dim_id) == len(bits[sl_id])
    npt.assert_array_equal(np.arange(d, dtype=int), list(itertools.chain(*dim_ids)))


def test_table(sublattices, dim_ids, dim_ids_table_active, dim_ids_table_all):
    # Assert include all sites.
    n_sites = sum(len(sublatt.sites) for sublatt in sublattices)
    n_codes = max(max(sublatt.encoding) for sublatt in sublattices) + 1
    npt.assert_array_equal(
        np.sort([site for sublatt in sublattices for site in sublatt.sites]),
        np.arange(n_sites, dtype=int),
    )
    assert dim_ids_table_all.shape == (n_sites, n_codes)
    assert dim_ids_table_active.shape == (n_sites, n_codes)
    for sublatt, dim_ids_sl in zip(sublattices, dim_ids):
        npt.assert_array_equal(
            dim_ids_table_all[sublatt.sites[:, None], sublatt.encoding]
            - np.array(dim_ids_sl, dtype=int)[None, :],
            0,
        )
        outer_codes = np.setdiff1d(
            np.arange(n_codes, dtype=int), sublatt.encoding
        ).astype(int)
        npt.assert_array_equal(
            dim_ids_table_all[sublatt.sites[:, None], outer_codes], -1
        )
        if not sublatt.is_active:
            npt.assert_array_equal(dim_ids_table_active[sublatt.sites, :], -1)
        if sublatt.is_active:
            npt.assert_array_equal(
                dim_ids_table_active[sublatt.active_sites[:, None], sublatt.encoding]
                - np.array(dim_ids_sl, dtype=int)[None, :],
                0,
            )
        npt.assert_array_equal(dim_ids_table_active[sublatt.restricted_sites, :], -1)


def test_occu_conversion(sublattices, dim_ids_table_all, dim_ids_table_active, rng):
    d = sum(len(sublatt.species) for sublatt in sublattices)
    for _ in range(10):
        occu = _gen_unconstrained_ordered_occu(sublattices, rng=rng)
        list_all = occu_to_species_list(occu, d, dim_ids_table_all)
        list_active = occu_to_species_list(occu, d, dim_ids_table_active)
        n_all = occu_to_counts(occu, d, dim_ids_table_all)
        n_active = occu_to_counts(occu, d, dim_ids_table_active)
        npt.assert_array_equal(n_all, [len(sites) for sites in list_all])
        npt.assert_array_equal(n_active, [len(sites) for sites in list_active])
        dim_id = 0
        for sublatt in sublattices:
            for code in sublatt.encoding:
                sites_all = list_all[dim_id]
                sites_active = list_active[dim_id]
                assert np.all(site in sublatt.active_sites for site in sites_active)
                npt.assert_array_equal(occu[sites_all], code)
                npt.assert_array_equal(occu[sites_active], code)
                if not sublatt.is_active:
                    assert len(sites_active) == 0
                dim_id += 1


def test_delta_n(sublattices, dim_ids_table_active, rng):
    # Test a few good steps.
    d = sum(len(sublatt.species) for sublatt in sublattices)
    occu = _gen_unconstrained_ordered_occu(sublattices, rng=rng)
    all_sites = np.arange(len(occu), dtype=int)
    activeness = np.any(dim_ids_table_active >= 0, axis=-1)
    # An active site must have at least 2 species allowed.
    assert np.all(
        np.sum(dim_ids_table_active[all_sites[activeness], :] >= 0, axis=-1) >= 2
    )
    active_sites = all_sites[activeness]
    for _ in range(10):
        site = np.random.choice(active_sites)
        code_ori = occu[site]
        table_site = dim_ids_table_active[site, :]
        code_nex = np.random.choice(
            list(
                set(np.arange(len(table_site), dtype=int)[table_site >= 0]) - {code_ori}
            )
        )
        step1 = [(site, code_nex)]
        step2 = [(site, code_nex), (site, code_ori)]
        step3 = [(site, code_nex), (site, code_nex)]
        dim_id_ori = dim_ids_table_active[site, code_ori]
        dim_id_nex = dim_ids_table_active[site, code_nex]

        dn1 = delta_counts_from_step(occu, step1, d, dim_ids_table_active)
        dn1_std = np.zeros(len(dn1))
        dn1_std[dim_id_nex] = 1
        dn1_std[dim_id_ori] = -1
        dn2 = delta_counts_from_step(occu, step2, d, dim_ids_table_active)
        dn2_std = np.zeros(len(dn2))
        dn3 = delta_counts_from_step(occu, step3, d, dim_ids_table_active)
        npt.assert_array_equal(dn1, dn1_std)
        npt.assert_array_equal(dn2, dn2_std)
        npt.assert_array_equal(dn3, dn1_std)

    # Test bad steps.
    site = np.random.choice(active_sites)
    table_site = dim_ids_table_active[site, :]
    bad_codes = np.arange(len(table_site), dtype=int)[table_site < 0]
    if len(bad_codes) > 0:
        bad_code = np.random.choice(bad_codes)
        bad_step = [(site, bad_code)]
        with pytest.raises(ValueError):
            _ = delta_counts_from_step(occu, bad_step, d, dim_ids_table_active)
    inactive_sites = np.setdiff1d(all_sites, active_sites)
    bad_site = np.random.choice(inactive_sites)
    rand_code = np.random.choice(np.arange(len(table_site), dtype=int))
    bad_step = [(bad_site, rand_code)]
    with pytest.raises(ValueError):
        _ = delta_counts_from_step(occu, bad_step, d, dim_ids_table_active)



"""Test boolean variables' generation for upper-bound groundstate."""

import numpy as np
import numpy.testing as npt
import pytest

from smol.capp.generate.groundstate.upper_bound.indices import (
    get_sublattice_indices_by_site,
)
from smol.capp.generate.groundstate.upper_bound.variables import (
    get_occupancy_from_variables,
    get_variable_values_from_occupancy,
    get_variables_from_sublattices,
)
from smol.cofe.space.domain import get_allowed_species

from .utils import get_random_variable_values


def test_upper_variables_and_indices(
    solver_test_ensemble, solver_test_initial_occupancy
):
    processor_structure = solver_test_ensemble.processor.structure
    orig_site_spaces = get_allowed_species(processor_structure)
    # If not given initial occupancy, will throw an error.
    with pytest.raises(ValueError):
        _ = get_variables_from_sublattices(
            solver_test_ensemble.sublattices, processor_structure
        )
    variables, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        processor_structure,
        solver_test_initial_occupancy,
    )

    assert len(variable_indices) == solver_test_ensemble.num_sites
    # Variable indices must be continuous.
    assert variables.size == max(max(site_vars) for site_vars in variable_indices) + 1
    flatten_indices = sorted(
        [ind for site_inds in variable_indices for ind in site_inds if ind >= 0]
    )
    assert list(range(variables.size)) == flatten_indices

    # Check for each site.
    for site_id, site_vars in enumerate(variable_indices):
        site_var_inds = np.array(site_vars, dtype=int)
        sublattice = None
        for sl in solver_test_ensemble.sublattices:
            if site_id in sl.sites:
                sublattice = sl
        # Contains inactive site or impossible species.
        if np.any(site_var_inds < 0):
            if np.any(site_var_inds >= 0):
                assert np.all(site_var_inds != -1)
            # Only one species can be fixed true.
            else:
                assert np.sum(site_var_inds == -1) == 1
                assert site_id in solver_test_ensemble.restricted_sites
                # And this fixed true species must match that in initial occupancy.
                # Inactive sub-lattice.
                if len(sublattice.species) == 1:
                    expected_sp_id = orig_site_spaces[site_id].index(
                        sublattice.species[0]
                    )
                # Manually restricted site.
                else:
                    expected_code = solver_test_initial_occupancy[site_id]
                    # Assume split from a fresh sub-lattice (code continuous from 0~n).
                    expected_sp_id = expected_code
                expected_site_vars = np.zeros(len(site_vars), dtype=int) - 2
                expected_site_vars[expected_sp_id] = -1
                npt.assert_array_equal(expected_site_vars, site_vars)
        else:
            assert len(site_vars) == len(sublattice.encoding)
            assert list(range(min(site_vars), max(site_vars) + 1)) == site_vars


def test_occupancy_from_variables(solver_test_ensemble, solver_test_initial_occupancy):
    _, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )

    site_sublattice_ids = get_sublattice_indices_by_site(
        solver_test_ensemble.sublattices
    )

    for _ in range(20):
        rand_vals = get_random_variable_values(solver_test_ensemble.sublattices)
        rand_occu = get_occupancy_from_variables(
            solver_test_ensemble.sublattices,
            rand_vals,
            variable_indices,
        )
        # Constrained sites must be the same.
        npt.assert_array_equal(
            rand_occu[solver_test_ensemble.restricted_sites],
            solver_test_initial_occupancy[solver_test_ensemble.restricted_sites],
        )
        # Check active sites match.
        for site_id in np.setdiff1d(
            np.arange(solver_test_ensemble.num_sites, dtype=int),
            solver_test_ensemble.restricted_sites,
        ):
            sublattice = solver_test_ensemble.sublattices[site_sublattice_ids[site_id]]
            expected_vals = np.array(sublattice.encoding == rand_occu[site_id]).astype(
                int
            )
            active_variable_indices = np.array(variable_indices[site_id], dtype=int)
            active_variable_indices = active_variable_indices[
                active_variable_indices >= 0
            ]
            input_vals = rand_vals[active_variable_indices].astype(int)
            npt.assert_array_equal(expected_vals, input_vals)


def test_variables_from_occupancy(solver_test_ensemble, solver_test_initial_occupancy):
    _, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    # Check conversions are correct for random variables.
    for _ in range(20):
        rand_val = get_random_variable_values(solver_test_ensemble.sublattices).astype(
            int
        )
        rand_occu = get_occupancy_from_variables(
            solver_test_ensemble.sublattices,
            rand_val,
            variable_indices,
        )

        test_val = get_variable_values_from_occupancy(
            solver_test_ensemble.sublattices, rand_occu, variable_indices
        ).astype(int)
        npt.assert_array_equal(rand_val, test_val)
        # print("Variable indices:\n", variable_indices)
        # print("rand_occu:\n", rand_occu)
        # print("rand_val:\n", rand_val)
        # assert False



"""smol -- Statistical Mechanics On Lattices"""

import os
import platform
import shutil
import sys

import numpy
from setuptools import Command, Extension, setup
from setuptools.command.build_ext import build_ext as build_ext_

from tools.build_helpers import (
    check_openmp_support,
    cythonize_extensions,
    get_openmp_flag,
)

COMPILE_OPTIONS = {
    "msvc": [
        "/O2",
        "/EHsc",
    ],
    "mingw32": ["-Wall", "-Wextra"],
    "other": ["-Wall", "-Wextra", "-ffast-math", "-O3"],
}

LINK_OPTIONS = {
    "msvc": ["-Wl, --allow-multiple-definition"],
    "mingw32": [["-Wl, --allow-multiple-definition"]],
    "other": [],
}

COMPILER_DIRECTIVES = {
    "language_level": 3,
}

if sys.platform.startswith("darwin"):
    if "PYTHON_CROSSENV" not in os.environ:
        COMPILE_OPTIONS["other"] += ["-mcpu=native"]
    elif platform.machine() == "x86_64":  # cross compiling for macos arm64
        COMPILE_OPTIONS["other"] += ["-mcpu=apple-m1"]
    # otherwise we don't specific -mcpu


# custom clean command to remove .c files
# taken from sklearn setup.py
class clean(Command):
    description = "Remove build artifacts from the source tree"

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Remove c files if we are not within a sdist package
        cwd = os.path.abspath(os.path.dirname(__file__))
        remove_c_files = not os.path.exists(os.path.join(cwd, "PKG-INFO"))
        if remove_c_files:
            print("Will remove generated .c files")
        if os.path.exists("build"):
            shutil.rmtree("build")
        for dirpath, dirnames, filenames in os.walk("smol"):
            for filename in filenames:
                _, extension = os.path.splitext(filename)

                if extension in [".so", ".pyd", ".dll", ".pyc"]:
                    os.unlink(os.path.join(dirpath, filename))

                if remove_c_files and extension in [".c"]:
                    pyx_file = str.replace(filename, extension, ".pyx")
                    if os.path.exists(os.path.join(dirpath, pyx_file)):
                        os.unlink(os.path.join(dirpath, filename))

            for dirname in dirnames:
                if dirname == "__pycache__":
                    shutil.rmtree(os.path.join(dirpath, dirname))


class build_ext_options:
    def build_options(self):
        OMP_SUPPORTED = check_openmp_support(self.compiler)
        for e in self.extensions:
            e.extra_compile_args += COMPILE_OPTIONS.get(
                self.compiler.compiler_type, COMPILE_OPTIONS["other"]
            )

            e.extra_link_args += LINK_OPTIONS.get(
                self.compiler.compiler_type, LINK_OPTIONS["other"]
            )
            if OMP_SUPPORTED:
                omp_flag = get_openmp_flag(self.compiler)
                e.extra_compile_args += omp_flag
                e.extra_link_args += omp_flag


class build_ext(build_ext_, build_ext_options):
    def build_extensions(self):
        build_ext_options.build_options(self)
        build_ext_.build_extensions(self)


# Compile option for cython extensions
cython_kwargs = {}
if "--force-cython" in sys.argv:
    cython_kwargs["force"] = True
    sys.argv.remove("--force-cython")
if "--annotate-cython" in sys.argv:
    cython_kwargs["annotate"] = True
    sys.argv.remove("--annotate-cython")

extensions = [
    Extension(
        "smol.utils.cluster.evaluator",
        ["smol/utils/cluster/evaluator.pyx"],
        language="c",
    ),
    Extension(
        "smol.utils.cluster.ewald",
        ["smol/utils/cluster/ewald.pyx"],
        language="c",
    ),
    Extension(
        "smol.utils.cluster.container",
        ["smol/utils/cluster/container.pyx"],
        language="c",
    ),
    Extension(
        "smol.utils.cluster.correlations",
        ["smol/utils/cluster/correlations.pyx"],
        language="c",
    ),
    Extension(
        "smol.utils._openmp_helpers",
        ["smol/utils/_openmp_helpers.pyx"],
        language="c",
    ),
]

cmdclass = {
    "clean": clean,
    "build_ext": build_ext,
}

setup(
    ext_modules=cythonize_extensions(extensions, **cython_kwargs),
    cmdclass=cmdclass,
    include_dirs=numpy.get_include(),
)



import numpy as np
import numpy.testing as npt
import pytest

from smol.cofe import ClusterExpansion
from smol.utils.cluster.correlations import (
    corr_from_occupancy,
    delta_corr_single_flip,
    delta_interactions_single_flip,
    interactions_from_occupancy,
)
from smol.utils.exceptions import StructureMatchError
from tests.utils import compute_cluster_interactions, gen_random_ordered_structure

ATOL = 2e4 * np.finfo(float).eps


# test the new evaluator vs the legacy functions
@pytest.mark.xfail(raises=StructureMatchError)
def test_corr_from_occu(cluster_subspace, supercell_matrix, rng):
    supercell_matrix = 2 * supercell_matrix

    # legacy orbit lists
    mappings = cluster_subspace.supercell_orbit_mappings(supercell_matrix)
    orbit_list = []
    for orbit, cluster_inds in zip(cluster_subspace.orbits, mappings):
        assert cluster_inds.dtype == np.int32
        assert orbit.flat_tensor_indices.dtype == np.int32
        assert orbit.flat_correlation_tensors.dtype == np.float64
        orbit_list.append(
            (
                orbit.bit_id,
                orbit.flat_tensor_indices,
                orbit.flat_correlation_tensors,
                cluster_inds,
            )
        )

    orbit_list_ratio = []
    for orbit, cluster_inds in zip(cluster_subspace.orbits, mappings):
        orbit_list_ratio.append(
            (
                orbit.bit_id,
                1.0,
                orbit.flat_tensor_indices,
                orbit.flat_correlation_tensors,
                cluster_inds,
            )
        )

    ratio = np.ones(len(cluster_subspace.orbits))
    indices = cluster_subspace.get_orbit_indices(supercell_matrix)

    for _ in range(10):
        structure = gen_random_ordered_structure(
            cluster_subspace.structure, supercell_matrix
        )

        # test correlations
        occu = cluster_subspace.occupancy_from_structure(
            structure, encode=True, scmatrix=supercell_matrix
        )
        assert occu.dtype == np.int32

        # legacy cython function
        corr = corr_from_occupancy(occu, len(cluster_subspace), orbit_list)

        npt.assert_allclose(
            corr,
            cluster_subspace.evaluator.correlations_from_occupancy(
                occu, indices.container
            ),
        )
        npt.assert_allclose(
            corr,
            cluster_subspace.corr_from_structure(structure, scmatrix=supercell_matrix),
        )

        # test correlation changes
        occu_f = occu.copy()
        occu_f[0] = 0 if occu[0] != 0 else 1
        occu_f[1] = 0 if occu[0] != 0 else 1
        corr_f = corr_from_occupancy(occu_f, len(cluster_subspace), orbit_list)

        npt.assert_allclose(
            delta_corr_single_flip(
                occu_f, occu, len(cluster_subspace), orbit_list_ratio
            ),
            cluster_subspace.evaluator.delta_correlations_from_occupancies(
                occu_f, occu, ratio, indices.container
            ),
        )

        # check that it also matches the full difference
        npt.assert_allclose(
            corr_f - corr,
            cluster_subspace.evaluator.delta_correlations_from_occupancies(
                occu_f, occu, ratio, indices.container
            ),
            atol=ATOL,
        )


def test_interactions_from_occu(cluster_subspace, supercell_matrix, rng):
    supercell_matrix = 2 * supercell_matrix
    eci = rng.random(len(cluster_subspace))
    expansion = ClusterExpansion(cluster_subspace, eci)
    flat_interaction_tensors = tuple(
        np.ravel(t, order="C") for t in expansion.cluster_interaction_tensors[1:]
    )
    offset = expansion.cluster_interaction_tensors[0]

    cluster_subspace.evaluator.set_cluster_interactions(
        flat_interaction_tensors,
        expansion.cluster_interaction_tensors[0],
    )
    # legacy orbit lists
    mappings = cluster_subspace.supercell_orbit_mappings(supercell_matrix)
    orbit_list = []
    for orbit, cluster_inds, flat_tensor in zip(
        cluster_subspace.orbits, mappings, flat_interaction_tensors
    ):
        orbit_list.append(
            (
                orbit.flat_tensor_indices,
                flat_tensor,
                cluster_inds,
            )
        )

    orbit_list_ratio = []
    for orbit, cluster_inds, flat_tensor in zip(
        cluster_subspace.orbits, mappings, flat_interaction_tensors
    ):
        orbit_list_ratio.append(
            (
                orbit.id,
                1.0,
                orbit.flat_tensor_indices,
                flat_tensor,
                cluster_inds,
            )
        )

    ratio = np.ones(len(cluster_subspace.orbits))
    indices = cluster_subspace.get_orbit_indices(supercell_matrix)

    for _ in range(10):
        structure = gen_random_ordered_structure(
            cluster_subspace.structure, supercell_matrix
        )

        # test correlations
        occu = cluster_subspace.occupancy_from_structure(
            structure, encode=True, scmatrix=supercell_matrix
        )

        # compare with explicit calculation of interactions
        # (see utils.compute_cluster_interactions)
        interactions = compute_cluster_interactions(
            expansion, structure, scmatrix=supercell_matrix
        )

        # legacy cython function
        npt.assert_allclose(
            interactions,
            interactions_from_occupancy(
                occu, cluster_subspace.num_orbits, offset, orbit_list
            ),
        )

        # calculations from cluster expansion
        npt.assert_allclose(
            interactions,
            expansion.cluster_interactions_from_structure(
                structure, scmatrix=supercell_matrix
            ),
        )

        # explicit calls to evaluator
        npt.assert_allclose(
            interactions,
            cluster_subspace.evaluator.interactions_from_occupancy(
                occu, indices.container
            ),
        )

        # test correlation changes
        occu_f = occu.copy()
        occu_f[0] = 0 if occu[0] != 0 else 1

        # legacy cython function
        npt.assert_allclose(
            delta_interactions_single_flip(
                occu_f, occu, cluster_subspace.num_orbits, orbit_list_ratio
            ),
            cluster_subspace.evaluator.delta_interactions_from_occupancies(
                occu_f, occu, ratio, indices.container
            ),
        )

        # check that it also matches a full difference with > 1 flips
        occu_f[1] = 0 if occu[0] != 0 else 1
        interactions_f = interactions_from_occupancy(
            occu_f, cluster_subspace.num_orbits, offset, orbit_list
        )

        npt.assert_allclose(
            interactions_f - interactions,
            cluster_subspace.evaluator.delta_interactions_from_occupancies(
                occu_f, occu, ratio, indices.container
            ),
            atol=ATOL,
        )



import numpy as np
import numpy.testing as npt
import pytest

from smol.capp.generate import StochasticSQSGenerator
from smol.utils.math import integerize_vector
from tests.utils import assert_pickles


@pytest.fixture(params=["correlation", "cluster-interaction"], scope="package")
def generator(structure, cluster_cutoffs, request):
    _, supercell_size = integerize_vector(
        [val for val in structure.composition.values()]
    )
    generator = StochasticSQSGenerator.from_structure(
        structure,
        cluster_cutoffs,
        supercell_size=supercell_size,
        feature_type=request.param,
    )
    return generator


def test_generate_get_sqs(generator):
    assert generator.num_structures == 0
    generator.generate(1000, max_save_num=100)
    assert 0 < generator.num_structures <= 100

    traces = list(generator._sqs_deque)
    assert all(
        tr1.enthalpy >= tr2.enthalpy for tr1, tr2 in zip(traces[:-1], traces[1:])
    )

    assert len(generator.get_best_sqs()) == 1
    assert (
        len(generator.get_best_sqs(generator.num_structures, remove_duplicates=False))
        == generator.num_structures
    )
    assert (
        len(generator.get_best_sqs(generator.num_structures))
        <= generator.num_structures
    )
    assert (
        len(generator.get_best_sqs(generator.num_structures, reduction_algorithm="LLL"))
        <= generator.num_structures
    )


def test_compute_score_distances(generator):
    generator.generate(1000, max_save_num=100)
    for sqs in generator.get_best_sqs(
        generator.num_structures, remove_duplicates=False
    ):
        assert generator.compute_score(
            sqs.structure, sqs.supercell_matrix
        ) == pytest.approx(sqs.score)

        npt.assert_allclose(
            generator.compute_feature_distance(sqs.structure, sqs.supercell_matrix),
            sqs.feature_distance,
            rtol=1e-05,
            atol=1e-08,  # np.allclose defaults
        )

    with pytest.raises(ValueError):
        size = generator.supercell_size + 1
        generator.compute_score(sqs.structure, size * np.eye(3))


def test_bad_generator(cluster_subspace):
    with pytest.raises(ValueError):
        StochasticSQSGenerator(cluster_subspace, 2, feature_type="blah")

    with pytest.raises(ValueError):
        StochasticSQSGenerator(
            cluster_subspace, 2, target_vector=np.ones(len(cluster_subspace) - 2)
        )

    with pytest.raises(ValueError):
        StochasticSQSGenerator(
            cluster_subspace, 2, target_weights=np.ones(len(cluster_subspace) + 2)
        )

    with pytest.raises(ValueError):
        StochasticSQSGenerator(
            cluster_subspace, 2, supercell_matrices=[np.ones((4, 3))]
        )

    with pytest.raises(ValueError):
        StochasticSQSGenerator(
            cluster_subspace, 2, supercell_matrices=[np.random.random((3, 3))]
        )


@pytest.mark.parametrize("feature_type", ["correlation", "cluster-interaction"])
def test_from_processors(single_subspace, feature_type):
    generator = StochasticSQSGenerator(
        single_subspace,
        supercell_size=2,
        feature_type=feature_type,
    )

    assert len(generator.processors) > 0

    new_generator = StochasticSQSGenerator.from_processors(generator.processors)

    assert (
        generator._processors_by_scm.keys() == new_generator._processors_by_scm.keys()
    )

    with pytest.raises(ValueError):
        StochasticSQSGenerator(
            generator.processors[0].cluster_subspace,
            supercell_size=3,
            processors=generator.processors,
        )

    with pytest.raises(ValueError):
        StochasticSQSGenerator(
            generator.processors[0].cluster_subspace,
            supercell_size=3,
            supercell_matrices=[np.eye(3)],
            processors=generator.processors,
        )


def test_pickles(generator):
    assert_pickles(generator)



"""Mathematic utilities.

Including linear algebra, combinatorics and integer enumerations.
"""

__author___ = "Fengyu Xie, Luis Barroso-Luque"

from fractions import Fraction
from itertools import combinations, product

import numpy as np
from monty.dev import requires
from scipy.linalg import null_space
from scipy.spatial import KDTree

try:
    import polytope as pc
except ImportError:
    pc = None

try:
    import cvxpy as cp
except ImportError:
    cp = None


# Global numerical tolerance in this module.
NUM_TOL = 1e-6


def yield_hermite_normal_forms(determinant):
    """Yield all hermite normal form matrices with given determinant.

    Args:
        determinant (int):
            determinant of hermite normal forms to be yielded

    Yields:
        ndarray: hermite normal form matrix with given determinant
    """
    for a in filter(lambda x: determinant % x == 0, range(1, determinant + 1)):
        quotient = determinant // a
        for c in filter(lambda x: quotient % x == 0, range(1, determinant // a + 1)):
            f = quotient // c
            for b, d, e in product(range(0, c), range(0, f), range(0, f)):
                yield np.array([[a, 0, 0], [b, c, 0], [d, e, f]], dtype=int)


def gcdex(a, b):
    """Extend Euclidean Algorithm."""
    if a == 0:
        return 0, 1, b

    x1, y1, g = gcdex(b % a, a)
    x = y1 - (b // a) * x1
    y = x1

    return x, y, g


def rationalize_number(a, max_denominator=1000, dtol=NUM_TOL):
    """Find a rational number near real number within dtol.

    Args:
        a(float):
            A number to be rationalized
        max_denominator(int): optional
            Maximum allowed denominator. Default 1000.
        dtol(float): optional
            Maximum allowed difference of variable a
            to its rational form. Default 1E-6.
            You must have 1/max_denominator > dtol!
    Return:
        Numerator and denominator (always positive):
           int, int
    """
    f = Fraction.from_float(a).limit_denominator(max_denominator)
    numerator = f.numerator
    denominator = f.denominator

    if abs(float(numerator) / denominator - a) > dtol:
        raise ValueError(
            "Can't find a rational number near " "{} within tolerance!".format(a)
        )

    return numerator, denominator


def integerize_vector(v, max_denominator=1000, dtol=NUM_TOL):
    """Integerize all components of a vector v.

    Rationalize all components of a vector, then multiply
    the vector by the LCM of all the rationalized component's
    denominator, so that all vector components are converted
    to integers. We call this process 'intergerization' of a
    vector.
    Args:
        v(np.ndarray(float)):
            A vector to be rationalized
        max_denominator(int,default=1000):
            Maximum allowed denominator.
        dtol(float,default=1E-6):
            Maximum allowed difference of variable a
            to its rational form.
        You must have 1/max_denominator > dtol!
    Return:
        Integrized vector, LCM of denominators:
            np.ndarray, int
    """
    denos = []
    v = np.array(v)
    for c in v:
        _, deno = rationalize_number(c, max_denominator=max_denominator, dtol=dtol)
        denos.append(deno)
    lcm = np.lcm.reduce(denos)
    return np.array(np.round(v * lcm), dtype=np.int64), lcm


def integerize_multiple(vs, max_denominator=1000, dtol=NUM_TOL):
    """Integerize multiple vectors in a matrix.

    Args:
        vs(np.ndarray(float)):
            A matrix of vectors to be rationalized.
        max_denominator(int,default=1000):
            Maximum allowed denominator.
        dtol(float,default=1E-6):
            Maximum allowed difference of variable to its rational form.
        You must have 1/max_denominator > dtol!
    Return:
        Integerized vectors in the input shape, LCM of denominator:
            np.ndarray[int], int
    """
    vs = np.array(vs)
    shp = vs.shape
    vs_flatten = vs.flatten()
    vs_flat_int, mul = integerize_vector(
        vs_flatten, max_denominator=max_denominator, dtol=dtol
    )
    vs_int = np.reshape(vs_flat_int, shp)
    return vs_int, mul


# Integer linear algebra utilities.
# Note: don't use sympy anymore. In sympy 1.9,
# a matrix m full of 0 will give is_zero=False,
# while in 1.5.1, it will give is_zero=True!
# Try to remove sympy dependency!
def compute_snf(a):
    """Compute smith normal form of a matrix.

    Args:
        a(2D arraylike of int):
            A matrix defined on some domain.

    Returns:
        s, m, t (np.ndarray of int):
            Smith decomposition of a, such that:
            m is the smith normal form and m = s a t.
    """

    def leftmult(m, i0, i1, a, b, c, d):  # Matrix pivoting operations.
        for j in range(m.shape[1]):
            x, y = m[i0, j], m[i1, j]
            m[i0, j] = a * x + b * y
            m[i1, j] = c * x + d * y

    def rightmult(m, j0, j1, a, b, c, d):  # Matrix pivoting operations.
        for i in range(m.shape[0]):
            x, y = m[i, j0], m[i, j1]
            m[i, j0] = a * x + c * y
            m[i, j1] = b * x + d * y

    # Must convert to a sympy.Matrix.
    m = np.round(a).astype(int).copy()
    s = np.eye(m.shape[0]).astype(int)
    t = np.eye(m.shape[1]).astype(int)
    last_j = -1
    for i in range(m.shape[0]):
        for j in range(last_j + 1, m.shape[1]):
            if not np.all(m[:, j] == 0):
                break
        else:
            break
        if m[i, j] == 0:
            for ii in range(m.shape[0]):
                if m[ii, j] != 0:
                    break
            leftmult(m, i, ii, 0, 1, 1, 0)
            leftmult(s, i, ii, 0, 1, 1, 0)
        rightmult(m, j, i, 0, 1, 1, 0)
        rightmult(t, j, i, 0, 1, 1, 0)
        j = i
        upd = True
        while upd:
            upd = False
            for ii in range(i + 1, m.shape[0]):
                if m[ii, j] == 0:
                    continue
                upd = True
                if m[ii, j] % m[i, j] != 0:
                    coef1, coef2, g = gcdex(m[i, j], m[ii, j])
                    coef3 = m[ii, j] // g
                    coef4 = m[i, j] // g
                    leftmult(m, i, ii, coef1, coef2, -coef3, coef4)
                    leftmult(s, i, ii, coef1, coef2, -coef3, coef4)
                coef5 = m[ii, j] // m[i, j]
                leftmult(m, i, ii, 1, 0, -coef5, 1)
                leftmult(s, i, ii, 1, 0, -coef5, 1)
            for jj in range(j + 1, m.shape[1]):
                if m[i, jj] == 0:
                    continue
                upd = True
                if m[i, jj] % m[i, j] != 0:
                    coef1, coef2, g = gcdex(m[i, j], int(m[i, jj]))
                    coef3 = m[i, jj] // g
                    coef4 = m[i, j] // g
                    rightmult(m, j, jj, coef1, -coef3, coef2, coef4)
                    rightmult(t, j, jj, coef1, -coef3, coef2, coef4)
                coef5 = m[i, jj] // m[i, j]
                rightmult(m, j, jj, 1, -coef5, 0, 1)
                rightmult(t, j, jj, 1, -coef5, 0, 1)
        last_j = j

    for i1 in range(min(m.shape)):
        for i0 in reversed(range(i1)):
            coef1, coef2, g = gcdex(m[i0, i0], m[i1, i1])
            if g == 0:
                continue
            coef3 = m[i1, i1] // g
            coef4 = m[i0, i0] // g
            leftmult(m, i0, i1, 1, coef2, coef3, coef2 * coef3 - 1)
            leftmult(s, i0, i1, 1, coef2, coef3, coef2 * coef3 - 1)
            rightmult(m, i0, i1, coef1, 1 - coef1 * coef4, 1, -coef4)
            rightmult(t, i0, i1, coef1, 1 - coef1 * coef4, 1, -coef4)

    s = np.array(s).astype(int)
    m = np.array(m).astype(int)
    t = np.array(t).astype(int)
    return s, m, t


def solve_diophantines(A, b=None):
    """Solve diophantine equations An=b.

    We use Smith normal form to solve equations. If equation is not solvable, we will
    throw an error.

    Note: If A decomposes to snf with large matrix elements, the numerical accuracy
    might have an issue! When this is the case, even if An=b has an integer solution,
    our function is not guaranteed to find it! But for most application uses, the
    numerical accuracy here should be enough.

    Args:
        A (2D ArrayLike[int]):
            Matrix A in An=b.
        b (1D ArrayLike[int], default=None):
            Vector b in An=b.
            If not given, will set to zeros.

    Return: (1D np.ndarray[int], 2D np.ndarray[int]
        A base solution and base vectors (as rows):
    """
    A = np.array(A, dtype=int)
    n, d = A.shape
    b = np.array(b, dtype=int) if b is not None else np.zeros(d, dtype=int)
    # If you choose b=0, the equations may not have
    # any natural number solution!

    U, B, V = compute_snf(A)
    c = U @ b
    k = None
    for i in range(min(n, d)):
        if B[i, i] == 0:
            k = i
    k = min(n, d) if k is None else k

    # Check feasibility
    for i in range(k):
        if c[i] % B[i, i] != 0:
            print("index:", i)
            print("c[i]:", c[i])
            print("b[ii]:", B[i, i])
            print("U:", U)
            print("B:", B)
            print("V:", V)
            assert np.allclose(B, U @ A @ V)
            raise ValueError("Diophantine equations A n = b are not feasible!")

    # Get base solution
    n0 = V[:, :k] @ (c[:k] // B.diagonal()[:k])

    return n0, V[:, k:].transpose().copy()


@requires(pc is not None, " 'polytope' package not found. Please install it.")
def get_nonneg_float_vertices(A, b):
    """Get vertices of polytope An=b, n>=0.

    n is allowed to be non-negative floats.
    This result can be used to get the minimum
    supercell size by looking for the minimum
    integer than can integerize all these vertices,
    if b is written according to number of sites in
    a primitive cell.

    Args:
        A (2D np.ndarray):
            A in An=b.
        b (2D np.ndarray):
            b in An=b.
    Returns:
        Vertices of polytope An=b, n>=0 in float:
            np.ndarray[float]
    """
    A = np.array(A)
    b = np.array(b)

    vs = null_space(A).transpose()  # each basis vector as a row.
    n0 = np.linalg.pinv(A) @ b
    poly = pc.Polytope(-1 * vs.transpose(), n0)

    verts = pc.extreme(poly)
    if len(verts) == 0:
        if pc.is_empty(poly):
            raise ValueError("Provided equation An=b is not feasible " "under n>=0.")
        else:
            raise ValueError(
                "Provided equation An=b is not fill " "dimensional under n>=0."
            )
    verts = verts @ vs + n0
    return verts


@requires(
    pc is not None and cp is not None,
    "'polytope' and 'cvxpy' packages are required. Please install them.",
)
def get_natural_centroid(
    n0,
    vs,
    sc_size,
    a_leq=None,
    b_leq=None,
):
    """Get the natural number solution closest to centroid.

    Done by linear programming, minimize:
    norm_2(x - centroid)**2
    s.t. n0 + sum_s x_s * v_s >= 0
    Where centroid is the center of the polytope
    bounded by those constraints.

    Note: Need cvxopt and cvxpy!

    Args:
        n0 (1D ArrayLike[int]):
            An origin point of integer lattice.
        vs (2D ArrayLike[int]):
            Basis vectors of integer lattice.
        sc_size (int):
            Super-cell size with n0 as a base solution.
        a_leq (2D ArrayLike[int]):
            Left-hand side of constraint A @ n <= b. Unit is per prim.
        b_leq(1D ArrayLike[float]):
            Right-hand side of constraint A @ n <= b. Unit is per prim.

    Returns: 1D np.ndarray[int]
        The natural number point on the grid closest to centroid ("x"):
    """
    n0 = np.array(n0, dtype=int)
    vs = np.array(vs, dtype=int)
    n, d = vs.shape
    assert len(n0) == d
    poly = pc.Polytope(-vs.transpose(), n0)
    centroid = np.average(pc.extreme(poly), axis=0)
    x = cp.Variable(n, integer=True)
    constraints = [n0[i] + vs[:, i] @ x >= 0 for i in range(d)]
    if a_leq is not None and b_leq is not None:
        for a, bb in zip(a_leq, b_leq):
            constraints.append(a @ (n0 + x @ vs) <= bb * sc_size)
    prob = cp.Problem(cp.Minimize(cp.sum_squares(x - centroid)), constraints)
    # Use gurobi if present.
    if "GUROBI" in cp.installed_solvers():
        _ = prob.solve(solver=cp.GUROBI)
    else:
        _ = prob.solve(
            solver=cp.ECOS_BB, max_iters=200, abstol=NUM_TOL, feastol=NUM_TOL
        )
    if x.value is None:
        raise ValueError("No feasible natural number composition found!")

    return np.array(np.round(x.value), dtype=int)


def get_one_dim_solutions(n0, v, integer_tol=NUM_TOL, step=1):
    """Solve one dimensional integer inequalities.

    This will solve n0 + v * x >= 0, give all
    integer solutions. x should be a single int.
    Args:
        n0 (1D ArrayLinke[int]:
            1 dimensional constraining factors
        v (1D ArrayLinke[int]:
            1 dimensional constraining factors
        integer_tol (float): optional
            Tolerance of a number off integer
            value. If (number - round(number))
            <= integer_tol, it will be considered
            as an integer. Default is set by global
            NUM_TOL.
        step (int): optional
            Step to skip when yielding solutions. For example,
            when step=2, will yield every 2 solutions.
            Default is 1, yield every solution.
    Returns:
        All Integer solutions:
            1D np.ndarray[int]
    """
    x_min = -np.inf
    x_max = np.inf
    for i in range(len(v)):
        if v[i] > 0:
            l_margin = -n0[i] / v[i]
            if l_margin > x_min:
                x_min = l_margin
        elif v[i] < 0:
            r_margin = -n0[i] / v[i]
            if r_margin < x_max:
                x_max = r_margin
        else:
            if n0[i] < 0:  # Dim i not feasible.
                return np.array([], dtype=int)

    if x_min <= -np.inf or x_max >= np.inf:
        raise ValueError("Inequalities are not bounded!")
    # If close to an integer, shift to it.
    # This is to prevent error in floor and ceil.
    x_min = round(x_min) if abs(x_min - round(x_min)) <= integer_tol else x_min
    x_max = round(x_max) if abs(x_max - round(x_max)) <= integer_tol else x_max

    n_min = np.ceil(x_min)
    n_max = np.floor(x_max)
    if n_min > n_max:
        return np.array([], dtype=int)
    else:
        return np.arange(n_min, n_max + 1, step, dtype=int)


@requires(cp is not None, "'cvxpy' package is required. Please install it.")
def get_first_dim_extremes(a, b):
    """Solve extremes for x0 under ax<=b.

    Args:
        a, b (ArrayLike[int]):
            Constraints ax<=b. ax<=b must be feasible and
            bounded.
    Return:
        float, float:
            min x0 and max x0 when ax<=b is feasible.
    """
    a = np.array(a)
    b = np.array(b)
    n, d = a.shape
    if len(b) != n:
        raise ValueError(f"Constraint matrix {a} and " f"vector {b} does not match!")
    x1 = cp.Variable(d)
    prob1 = cp.Problem(cp.Minimize(x1[0]), [a @ x1 <= b])
    _ = prob1.solve()
    if x1.value is None:
        raise ValueError(f"Polytope a: {a}, b:{b} is empty or not bounded!")
    assert abs(prob1.value - x1.value[0]) <= NUM_TOL

    x2 = cp.Variable(d)
    prob2 = cp.Problem(cp.Maximize(x2[0]), [a @ x2 <= b])
    _ = prob2.solve()
    if x1.value is None:
        raise ValueError(f"Polytope a: {a}, b:{b} is empty or not bounded!")
    assert abs(prob2.value - x2.value[0]) <= NUM_TOL

    return prob1.value, prob2.value


def get_natural_solutions(n0, vs, integer_tol=NUM_TOL, step=1):
    """Enumerate all natural number solutions.

    Given the convex polytope n0 + sum_i x_i*v_i >= 0.
    Each time enumerate all possible integer x_0,
    fix x_0, then solve in each polytope with 1
    less dimension:
    n0 - x_0 * v_0 + sum_(i>0) x_i*v_i >= 0
    Recurse until all solutions are enumerated.

    Notice:
        1, This function is very costly! Do not
    use it with a large super-cell size!
        2, This function does not apply to any n0
    and vs. It only applies to bounded polytopes!

    Args:
        n0(1D ArrayLike[int]):
            An origin point of integer lattice.
        vs(2D ArrayLike[int]):
            Basis vectors of integer lattice.
        integer_tol(float): optional
            Tolerance of a number off integer
            value. If (number - round(number))
            <= integer_tol, it will be considered
            as an integer. Default is set by global
            NUM_TOL.
        step(int): optional
            Step to skip when yielding solutions in each
            dimension. For example,
            when step=2, will yield every 2 solutions in each
            dimension.
            Default is 1, yield every solution.

    Returns:
        All natural number solutions ("x"):
            2D np.ndarray[int]
    """
    n0 = np.array(n0, dtype=int)
    vs = np.array(vs, dtype=int)

    n, d = vs.shape
    if n == 1:
        sols = get_one_dim_solutions(n0, vs[0, :], integer_tol=integer_tol, step=step)
        sols = sols.reshape(-1, 1)

        return sols

    x_min, x_max = get_first_dim_extremes(-1 * vs.transpose(), n0)
    if x_min <= -np.inf or x_max >= np.inf:
        raise ValueError("Inequalities are not bounded!")
    # Do not use polytope module here. A polytope cannot be take extreme
    # if it is not feasible or has 0 volume.
    # Always branch the 1st dimension.

    x_min = round(x_min) if abs(x_min - round(x_min)) <= integer_tol else x_min
    x_max = round(x_max) if abs(x_max - round(x_max)) <= integer_tol else x_max

    n_min = np.ceil(x_min)
    n_max = np.floor(x_max)
    if n_min > n_max:
        return np.array([], dtype=int).reshape(-1, n)
    else:
        n_range = np.arange(n_min, n_max + 1, step, dtype=int)
        sols = []
        for m in n_range:
            n0_next = m * vs[0, :] + n0
            vs_next = vs[1:, :]
            sols_m = get_natural_solutions(
                n0_next, vs_next, integer_tol=integer_tol, step=step
            )
            if len(sols_m) > 0:
                sols_m = np.append(
                    m * np.ones(len(sols_m), dtype=int).reshape(-1, 1), sols_m, axis=-1
                )
            else:
                sols_m = np.array([], dtype=int).reshape(-1, n)
            sols.append(sols_m)

        return np.concatenate(sols, axis=0)


# Flip table utilities
def flip_size(u):
    """Get metric of a direction.

    This function can compute flip size, absolute
    charge on one side of the reaction formula, etc.
    Args:
        u(1D ArrayLike[int]):
            A flip direction on the composition lattice.
            The components of u must be ordered, such that
            u is a simple concatenation of components on
            each sub-lattice.
            We only check that sum(u) == 0. It is your
            responsibility to check that's also true
            per-sub-lattice.

    Returns:
        Metric of direction u:
            int
    """
    u = np.array(u, dtype=int)
    if np.sum(u) != 0:
        raise ValueError(f"Flip vector {u} does not" "conserve number of sites!")
    return np.sum(u[u > 0])


def count_row_matches(a1, a2):
    """Count number of matching rows in integer arrays.

    Args:
        a1, a2(2D Arraylike[int]):
            Arrays to compare
    Returns:
        Number of rows in common:
           int
    """
    a1 = np.array(a1, dtype=int)
    a2 = np.array(a2, dtype=int)
    s1 = {tuple(r) for r in a1}
    s2 = {tuple(r) for r in a2}
    return len(s1 & s2)


def connectivity(u, ns):
    """Compute graph connectivity contributed by vector.

    connectivity = number of (n',n) pair in natural
    number solutions ns, that n'-n = u or -u.
    i.e., suppose the natural number solutions as
    fully connected graph nodes, connectivity contributed
    by u is the number of edges parallel to u.

    Args:
        u(1D ArrayLike[int]):
            A flip direction on the composition lattice.
            The components of u must be ordered, such that
            u is a simple concatenation of components in
            each sub-lattice.
        ns(2D ArrayLike[int]):
            A grid of natural number solutions.

    Returns:
        connectivity contributed by u: int
    """
    u = np.array(u, dtype=int)
    ns = np.array(ns, dtype=int)

    # Connectivity of u must be the same as -u,
    # This is trivial conclusion from sets theory,
    # so do not double-count.
    return count_row_matches(ns, ns + u)


def is_connected(n, vs, ns):
    """Check whether a grid point is connected by flip vectors.

    Args:
        n(1D ArrayLike[int]):
            A grid point.
        vs(2D ArrayLike[int]):
            Table of flip vectors.
        ns(2D ArrayLike[int]):
            All grid points.
    Returns:
        bool.
    """
    n = np.array(n, dtype=int)
    vs = np.array(vs, dtype=int)
    ns = np.array(ns, dtype=int)
    n_images = np.concatenate((vs, -vs), axis=0) + n
    return np.any(np.all(np.isclose(n_images[:, None, :], ns[None, :, :]), axis=-1))


def get_optimal_basis(n0, vs, xs, max_loops=100):
    """Get the optimal basis vectors to include in the flip table.

    Generate optimal basis vectors by:
        (1) Requiring basis vectors have minimal flip sizes;
        (2) Requiring maximum connectivity between compositions.
    To do this, we will iteratively execute the following steps
    to do a greedy optimization:
        1, For the current basis v1, .., vn, enumerate vectors:
        {v1, ..., vn} + {vi + vj | (i, j) in choose(n, 2)} +
        {vi - vj | (i, j) in choose(n, 2)}
        2, Sort these vectors, primarily by ascending flip size,
        secondarily by descending contribution to connectivity.
        3, Create an empty list. Iterate over sorted vectors.
        If the current we are looking at is not linearly dependent
        on the vectors already in the list, append it to the list.
        Stop the iteration until the list has n vectors.
    The above is called a cycle of optimization. We loop cycles
    to iteratively optimize basis set, until max number of loops
    has been reached, or the selected basis vectors no longer change.

    Notice:
        (1) This function is very costly! Do not use it with a large
    super-cell size!
        (2) The algorithm is supposed to give relatively good basis
    vectors, but its completeness has not been proven. Use it at your
    own risk.

    Args:
        n0(1D ArrayLike[int]):
            An origin point of integer lattice.
        vs(2D ArrayLike[int]):
            Basis vectors of integer lattice.
        xs(2D ArrayLike[int]):
            Solutions such that for each row x in xs,
            n0.T + x.T @ vs is a natural number point
            on the lattice.
            Shape = (n_solutions, n_basis_vectors)
        max_loops(int): optional
            Maximum number of optimization loops.

    Returns:
        Optimized basis vectors:
            2D np.ndarray[int]
    """
    n0 = np.array(n0, dtype=int)
    vs = np.array(vs, dtype=int)
    vs_opt = vs.copy()
    xs = np.array(xs, dtype=int)

    ns = xs @ vs + n0
    n, d = vs.shape

    def key_func(u):
        return flip_size(u), -1 * connectivity(u, ns)

    # Make the first column always positive.
    def standardize_table(V):
        sign = (V[:, 0] >= 0).astype(int)
        mult = np.round((sign - 0.5) * 2).astype(int)
        return V * mult[:, None]

    def table_match(V1, V2):
        if V1.shape != V2.shape:
            return False
        V1 = standardize_table(V1)
        V2 = standardize_table(V2)
        return count_row_matches(V1, V2) == V1.shape[0]

    for _ in range(max_loops):
        V = vs_opt.copy()
        for i1, i2 in combinations(list(range(n)), 2):
            V = np.concatenate((V, [V[i1] + V[i2], V[i1] - V[i2]]), axis=0)

        V = np.array(sorted(V, key=key_func), dtype=int)
        vs_new = np.array([], dtype=int).reshape(0, d)
        for i in range(len(V)):
            if len(vs_new) == n:
                break
            vs_current = np.concatenate((vs_new, [V[i]]), axis=0)
            if np.linalg.matrix_rank(vs_current) == min(vs_current.shape):
                # Full rank.
                vs_new = vs_current.copy()

        if table_match(vs_new, vs_opt):
            break
        vs_opt = vs_new.copy()

    return vs_opt


def get_ergodic_vectors(n0, vs, xs, k=3):
    """Compute an ergodic flip table.

    Notice:
        1, The following algorithm only guarantees getting
        an ergodic flip table.
        It will also try to prioritize adding flip directions
        with  minimal flip sizes, but global optimality is not
        guaranteed. Currently also does not guarantee graph
        connectivity, if graphs can be divided as parts.
        2, This algorithm does not guarantee finding the
        fewest number of directions to satisfy ergodicity at all.
        2, This process is NP-hard. Do not use it when the
        super-cell size is large.
    Args:
        n0(1D ArrayLike[int]):
            An origin point of integer lattice.
        vs(2D ArrayLike[int]):
            Basis vectors of integer lattice.
        xs(2D ArrayLike[int]):
            Solutions such that for each row x in xs,
            n0.T + x.T @ vs is a natural number point
            on the lattice.
            Shape = (n_solutions, n_basis_vectors)
        k(int): optional
            Find k-nearest neighbor of non-ergodic points,
            add those of them with minimal flip size to ensure
            ergodicity. Default is 3. k>1 is required, otherwise
            will always return self.

    Returns:
        Basis + Flip directions added to guarantee ergodicity:
            2D np.ndarray[int]
    """

    def test_connected(vs, ns_disconnected, ns):
        return np.array([is_connected(n, vs, ns) for n in ns_disconnected], dtype=bool)

    def candidate_key(u):
        return flip_size(u)

    n0 = np.array(n0, dtype=int)
    xs = np.array(xs, dtype=int)
    vs = np.array(vs, dtype=int)
    ns = xs @ vs + n0
    connected = test_connected(vs, ns, ns)
    ns_disconnected = ns[~connected]
    if len(ns_disconnected) == 0:
        return vs

    tree = KDTree(ns)
    candidate_vectors = []
    # Get k nearest neighbors in Euclidean distance.
    for n in ns_disconnected:
        dists, point_ids = tree.query(n, k=k)
        if dists[0] == 0:
            point_ids = point_ids[1:]
        points = ns[point_ids, :]
        for point in points:
            u = point - n
            if (
                tuple(u.tolist()) in candidate_vectors
                or tuple((-u).tolist()) in candidate_vectors
            ):
                continue
            candidate_vectors.append(tuple(u.tolist()))

    candidate_vectors = sorted(candidate_vectors, key=candidate_key)
    candidate_vectors = np.array(candidate_vectors, dtype=int)
    selected_vectors = vs.copy()
    ns_rem = ns_disconnected.copy()
    for u in candidate_vectors:
        vs_current = np.concatenate((selected_vectors, [u]), axis=0)
        connected = test_connected(vs_current, ns_rem, ns)
        selected_vectors = vs_current.copy()
        ns_rem = ns_rem[~connected]
        if len(ns_rem) == 0:
            break

    return selected_vectors


def flip_weights_mask(flip_vectors, n, max_n=None):
    """Mark feasibility of flip vectors.

    If a flip direction leads to any n+v < 0, then it is marked
    infeasible. Generates a boolean mask, every two components
    marks whether a flip direction and its inverse is feasible
    given n at the current occupancy.
    Will be used by Tableflipper.

    Args:
        flip_vectors(1D ArrayLike[int]):
            Flip directions in the table (inverses not included).
        n(1D ArrayLike[int]):
            Amount of each specie on sublattices. Same as returned
            by occu_to_species_n.
        max_n(1D ArrayLike[int]): optional
            Maximum number of each species allowed. This is needed
            When the number of active sites != number of sites in
            some sub-lattice.

    Return:
        Direction and its inverse are feasible or not:
           1D np.ndarray[bool]
    """
    flip_vectors = np.array(flip_vectors, dtype=int)
    directions = np.concatenate([(u, -u) for u in flip_vectors], axis=0)
    if max_n is None:
        max_n = np.ones(len(n)) * np.inf
    elif isinstance(max_n, (int, np.int32, np.int64)):
        max_n = np.ones(len(n), dtype=int) * max_n
    else:
        max_n = np.array(max_n, dtype=int)
    return ~(
        np.any(directions + n < 0, axis=-1) | np.any(directions + n > max_n, axis=-1)
    )


# Probability selection tools
def choose_section_from_partition(probabilities, rng=None):
    """Choose one partition from multiple partitions.

    This function choose one section from a list with probability weights.
    Args:
        probabilities(1D Arraylike[float]):
            Probabilities of each section. Will be normalized if not yet.
        rng (np.Generator): optional
            The given PRNG must be the same instance as that used by the
            kernel and any bias terms, otherwise reproducibility will be
            compromised.
    Return:
        The index of randomly chosen element:
           int
    """
    if rng is None:
        rng = np.random.default_rng()
    p = np.array(probabilities)
    if np.allclose(p, 0):
        p = np.ones(len(p))
    if not np.all(p >= -NUM_TOL):
        raise ValueError("Probabilities contain negative number.")
    p = p / p.sum()
    return int(round(rng.choice(len(p), p=p)))



"""Utility functions to handle encoded occupation arrays."""

__author__ = "Fengyu Xie"

import numpy as np


def get_dim_ids_by_sublattice(bits):
    """Get the component index of each species in vector n.

    Args:
        bits(List[List[Specie|Vacancy|Element]]):
           Species on each sub-lattice.
    Returns:
        Component index of each species on each sublattice in vector n:
           List[List[int]]
    """
    dim_ids = []
    dim_id = 0
    for species in bits:
        dim_ids.append(list(range(dim_id, dim_id + len(species))))
        dim_id += len(species)
    return dim_ids


# Parsing will be faster based on table.
def get_dim_ids_table(sublattices, active_only=False):
    """Get the dimension indices of all (site, code) in "couts" representation.

    This will be used to efficiently map occupancy to "couts" representation.
    Args:
        sublattices(smol.moca.Sublattice):
            All sub-lattices, active or not. The union
            of all sub-lattices' sites must be
            range(number of sites).
        active_only(bool): optional
            If true, will count un-restricted sites on all
            sub-lattices only. Default to false, will count
            all sites and sub-lattices.
    Returns:
        np.ndarray: a table storing the index of components corresponding to each
        site (rows) and each species on the site (columns).
    """
    n_row = sum(len(sublatt.sites) for sublatt in sublattices)
    n_col = max(max(sublatt.encoding) for sublatt in sublattices) + 1

    table = np.zeros((n_row, n_col), dtype=int) - 1
    dim_id = 0
    for sl_id, sublatt in enumerate(sublattices):
        for code in sublatt.encoding:
            if active_only:
                sites = sublatt.active_sites
            else:
                sites = sublatt.sites
            sites = sites.astype(int)  # in case sites are void.
            table[sites, code] = dim_id
            dim_id += 1
    return table


# Utilities for parsing occupation to composition.
def occu_to_species_list(occupancy, n_dims, dim_ids_table):
    """Get occupancy status of each sub-lattice.

    Get table of the indices of sites that are occupied by each specie on
    sub-lattices, from an encoded occupancy array.

    Args:
        occupancy(1d Arraylike[int]):
            An array representing encoded occupancy, can be list.
        n_dims (int):
            Number of dimensions in the unconstrained composition space,
            namely, the number of components in "counts" format vector
            representation.
        dim_ids_table(2D arrayLike[int]):
            Dimension indices of each site and code in n-representation.
            Rows correspond to site index, while columns correspond to
            species code. if a site is not active, all codes (columns)
            will give -1.

    Return:
        Index of sites occupied by each species, sublattices concatenated:
            List[List[int]]
    """
    occu = np.array(occupancy, dtype=int)
    if len(occu) != len(dim_ids_table):
        raise ValueError(
            f"Occupancy size {len(occu)} does not match "
            f"table size {len(dim_ids_table)}!"
        )
    dim_ids = dim_ids_table[np.arange(len(occu), dtype=int), occu]
    all_sites = np.arange(len(occu), dtype=int)
    return [all_sites[dim_ids == i].tolist() for i in range(n_dims)]


def occu_to_counts(occupancy, n_dims, dim_ids_table):
    """Count number of species from occupation array.

    Get the count of each species on sub-lattices from an encoded
    occupancy array. ("counts" format)
    Args:
        occupancy(1d Arraylike[int]):
            An array representing encoded occupancy, can be list.
        n_dims (int):
            Number of dimensions in the unconstrained composition space,
            namely, the number of components in "counts" format vector
            representation.
        dim_ids_table(2D arrayLike[int]):
            Dimension indices of each site and code in n-representation.
            Rows correspond to site index, while columns correspond to
            species code. if a site is not active, all codes (columns)
            will give -1.

    Return:
        Amount of each species, concatenated by sub-lattice:
            1D np.ndarray[int]
    """
    occu = np.array(occupancy, dtype=int)
    if len(occu) != len(dim_ids_table):
        raise ValueError(
            f"Occupancy size {len(occu)} does not match "
            f"table size {len(dim_ids_table)}!"
        )
    dim_ids = dim_ids_table[np.arange(len(occu), dtype=int), occu]
    n = np.zeros(n_dims, dtype=int)
    dims, counts = np.unique(dim_ids, return_counts=True)
    n[dims[dims >= 0]] = counts[dims >= 0]
    return n


def delta_counts_from_step(occu, step, n_dims, dim_ids_table):
    """Get the change of species amounts from MC step.

    Args:
        occu(1D arrayLike[int]):
            Encoded occupation array.
        step(List[tuple(int,int)]):
            List of tuples recording (site_id, code_of_species_to
            _replace_with).
        n_dims (int):
            Number of dimensions in the unconstrained composition space,
            namely, the number of components in "counts" format vector
            representation.
        dim_ids_table(2D arrayLike[int]):
            Dimension indices of each site and code in n-representation.
            Rows correspond to site index, while columns correspond to
            species code.

    Return:
        Change of species amounts (in "counts" format):
            1D np.ndarray[int]
    """
    occu_now = np.array(occu, dtype=int)
    dim_ids_table = np.array(dim_ids_table, dtype=int)
    delta_n = np.zeros(n_dims, dtype=int)
    # A step may involve a site multiple times.
    for site, code in step:
        code_ori = occu_now[site]
        dim_ori = dim_ids_table[site, code_ori]
        dim_nex = dim_ids_table[site, code]
        if dim_ori < 0 or dim_nex < 0:
            raise ValueError(
                f"Inactive sites or impossible codes " f"involved in step {step}!"
            )
        delta_n[dim_ori] -= 1
        delta_n[dim_nex] += 1
        occu_now[site] = code

    return delta_n



"""Simple mixin class that has an attributed/method that runs in parallel."""

import os
import warnings

from .._openmp_helpers import _openmp_effective_numthreads

if os.getenv("OMP_NUM_THREADS") is not None:
    DEFAULT_NUM_THREADS = _openmp_effective_numthreads()
else:
    DEFAULT_NUM_THREADS = _openmp_effective_numthreads(n_threads=2)


class SetNumThreads:
    """
    A descriptor used to set threads of an attributed object that has multi-threading.

    Right now it's only used for the Evaluator class.
    """

    def __init__(
        self, multithreaded_object_name: str, thread_attr_name: str = "num_threads"
    ):
        """Define the name of the multithread object and the threads attribute name."""
        # attribute that has multi-threading must have an int attribute
        # with name thread_attr_name
        self._obj_name = multithreaded_object_name
        self._attr_name = thread_attr_name

    def __get__(self, instance, objtype=None):
        """Get the number of threads used by the evaluator to compute correlations."""
        return getattr(getattr(instance, self._obj_name), self._attr_name)

    def __set__(self, instance, value):
        """Set the number of threads used by the evaluator to compute correlations."""
        if value is None:
            value = DEFAULT_NUM_THREADS

        if not isinstance(value, int):
            raise TypeError("num_threads must be an integer")

        max_threads = _openmp_effective_numthreads()
        if value > max_threads:
            warnings.warn(
                f"num_threads cannot be greater than {max_threads}. "
                f"Setting to {max_threads}."
                "If you want to use more threads, make sure openmp is enabled and set"
                "the OMP_NUM_THREADS environment variable accordingly."
            )
            value = max_threads

        obj = getattr(instance, self._obj_name)
        num_threads = _openmp_effective_numthreads(value)
        setattr(obj, self._attr_name, num_threads)
